# -*- coding: utf-8 -*-

# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER type system.

This module implements polymorphic function dispatch.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

import functools
import threading


class polymorphic(object):
    """Polymorphic function that dispatches on the type of the first arg.

    This function decorator can be used on instance methods as well as regular
    functions. It allows the function to dispatch on the type of its first
    argument, much like standard python instance methods dispatch on the type
    of self (conceptually, not in actuality).

    This enables us to define arbitrary interfaces and have already existing
    types participate in those interfaces, without having to actually alter
    the existing type hierarchy or monkey-patch additional functions into their
    namespaces.

    This approach is used in EFILTER to enable it to be easily added to
    existing codebases, which may already overload many operators and have
    their own conventions about how members of objects are accessed and types
    interact.

    Examples:
        @polymorphic
        def say_moo(bovine):
            raise NotImplementedError()

        class Cow():
            pass

        say_moo.implement(for_type=Cow, implementation=lambda x: "Moo!")

        class Sheep():
            pass

        say_moo.implement(for_type=Sheep, implementation=lambda x: "Baah!")

        shaun = Sheep()
        bessy = Cow()

        say_moo(shaun)  # => "Baah!"
        say_moo(bessy)  # => "Moo!"
    """

    # Locks _dispatch_table and implementations.
    _write_lock = None

    # Cache of type -> implementation.
    _dispatch_table = None

    # Table of which dispatch type is preferred over which other type in
    # cases that benefit from disambiguation.
    _prefer_table = None

    implementations = None
    func = None

    is_polymorphic_function = True

    def __init__(self, func):
        self._write_lock = threading.Lock()
        self.func = func
        self._dispatch_table = {}
        self._prefer_table = {}
        self.implementations = []
        functools.update_wrapper(self, func)

    @property
    def func_name(self):
        return self.func.func_name

    def __repr__(self):
        return "polymorphic(%s)" % self.func_name

    def __call__(self, obj, *args, **kwargs):
        implementation = self._find_and_cache_best_function(type(obj))
        if implementation:
            return implementation(obj, *args, **kwargs)

        # Fall-through to calling default implementation. By convention, the
        # default will usually raise a NotImplemented exception, but there
        # may be times when it will actually do something useful (good example
        # are convenience type checking functions, such as issuperposition).
        try:
            return self.func(obj, *args, **kwargs)
        except NotImplementedError:
            # Throw a better exception.
            raise NotImplementedError(
                "Polymorphic function %r has no concrete implementation for "
                "type %r and no default implementation. Available handlers: %r"
                % (self.func_name, type(obj), self.implementations))

    def implemented_for_type(self, dispatch_type):
        candidate = self._find_and_cache_best_function(dispatch_type)
        if candidate == self.func:
            return False

        return True

    def _preferred(self, preferred, over):
        prefs = self._prefer_table.get(preferred)
        if prefs and over in prefs:
            return True

        return False

    def prefer_type(self, prefer, over):
        """Prefer one type over another type, all else being equivalent.

        With abstract base classes (Python's abc module) it is possible for
        a type to appear to be a subclass of another type without the supertype
        appearing in the subtype's MRO. As such, the supertype has no order
        with respect to other supertypes, and this may lead to amguity if two
        implementations are provided for unrelated abstract types.

        In such cases, it is possible to disambiguate by explictly telling the
        function to prefer one type over the other.

        Arguments:
            prefer: Preferred type (class).
            over: The type we don't like (class).

        Raises:
            ValueError: In case of logical conflicts.
        """
        self._write_lock.acquire()
        try:
            if self._preferred(preferred=over, over=prefer):
                raise ValueError(
                    "Type %r is already preferred over %r." % (over, prefer))
            prefs = self._prefer_table.setdefault(prefer, set())
            prefs.add(over)
        finally:
            self._write_lock.release()

    def _find_and_cache_best_function(self, dispatch_type):
        """Finds the best implementation of this function given a type.

        This function caches the result, and uses locking for thread safety.

        Returns:
            Implementing function, in below order of preference:
            1. Explicitly registered implementations (through
               polymorphic.register) for types that arg_type either is
               or inherits from directly.
            2. Explicitly registered implementations accepting an abstract type
               (interface) in which dispatch_type participates (through
               abstract_type.register()).
            3. Default behavior of the polymorphic function. This will usually
               raise a NotImplementedError, by convention.

        Raises:
            TypeError: If two implementing functions are registered for
                different abstract types, and dispatch_type participates in
                both, and no order of preference was specified using
                prefer_type.
        """
        result = self._dispatch_table.get(dispatch_type)
        if result:
            return result

        self._write_lock.acquire()

        try:
            dispatch_mro = dispatch_type.mro()
        except TypeError:
            # Not every type has an MRO.
            dispatch_mro = ()

        best_match = None
        result_type = None
        try:
            for candidate_type, candidate_func in self.implementations:
                if not issubclass(dispatch_type, candidate_type):
                    # Skip implementations that are obviously unrelated.
                    continue

                try:
                    # The candidate implementation may be for a type that's
                    # actually in the MRO, or it may be for an abstract type.
                    match = dispatch_mro.index(candidate_type)
                except ValueError:
                    # This means we have an implementation for an abstract
                    # type, which ranks below all concrete types.
                    match = None

                if best_match is None:
                    if result and match is None:
                        # Already have a result, and no order of preference.
                        # This is probably because the type is a member of two
                        # abstract types and we have separate implementations
                        # for those two abstract types.

                        if self._preferred(candidate_type, over=result_type):
                            result = candidate_func
                            result_type = candidate_type
                        elif self._preferred(result_type, over=candidate_type):
                            # No need to update anything.
                            pass
                        else:
                            raise TypeError(
                                "Two candidate implementations found for "
                                "polymorphic function %s (dispatch type %s) "
                                "and neither is preferred." %
                                (self.func.func_name, dispatch_type))
                    else:
                        result = candidate_func
                        result_type = candidate_type

                if match < best_match:
                    result = candidate_func
                    result_type = candidate_type
                    best_match = match

            self._dispatch_table[dispatch_type] = result
            return result
        finally:
            self._write_lock.release()

    def implement(self, implementation, for_type=None, for_types=None):
        """Registers an implementing function for for_type.

        Arguments:
            implementation: Callable implementation for this type.
            for_type: The type this implementation applies to.
            for_types: Same as for_type, but takes a tuple of types.

            for_type and for_types cannot both be passed (for obvious reasons.)

        Raises:
            ValueError
        """
        if for_type:
            if for_types:
                raise ValueError("Cannot pass both for_type and for_types.")
            for_types = (for_type,)
        elif for_types:
            if not isinstance(for_types, tuple):
                raise TypeError("for_types must be passed as a tuple of "
                                "types (classes).")
        else:
            raise ValueError("Must pass either for_type or for_types.")

        for t in for_types:
            self._write_lock.acquire()
            try:
                self.implementations.append((t, implementation))
            finally:
                self._write_lock.release()



import unittest

from efilter.protocols import indexable
from efilter.ext import indexset


class FakeIndexable(object):
    def __init__(self, indices, value):
        self.indices = indices
        self.value = value

    def __repr__(self):
        return "FakeIndexable(%s)" % repr(self.value)


indexable.IIndexable.implement(
    for_type=FakeIndexable,
    implementations={
        indexable.indices: lambda obj: obj.indices
    }
)


class IndexSetTest(unittest.TestCase):
    def testSingleSet(self):
        e1 = FakeIndexable(["enum_foo", 1], "foo")
        e2 = FakeIndexable(["enum_bar", 2, "bar"], "bar")
        e3 = FakeIndexable(["enum_baz", 3], "baz")

        iset = indexset.IndexSet([e1, e3])

        self.assertItemsEqual(iset.values, [e1, e3])
        self.assertEqual(e1, iset.get_index(1))
        self.assertEqual(e1, iset.get_index("enum_foo"))

        self.assertEqual(len(iset), 2)
        iset.add(e3)
        self.assertEqual(len(iset), 2)
        iset.add(e2)
        self.assertEqual(len(iset), 3)
        iset.remove(e1)
        self.assertEqual(len(iset), 2)

        with self.assertRaises(KeyError):
            iset.remove(e1)

        self.assertEqual(len(iset), 2)
        self.assertItemsEqual(iset.values, [e2, e3])

        iset.remove(e2)
        iset.discard(e1)
        iset.pop()
        self.assertEqual(len(iset), 0)
        self.assertEqual(bool(iset), False)
        self.assertEqual(iset.values, [])

        iset.add(e1)
        self.assertEqual(iset.pop().value, e1.value)

    def testSetUnion(self):
        elements = [FakeIndexable([i, "s%d" % i, (i, None)], i)
                    for i in xrange(20)]

        iset1 = indexset.IndexSet(elements[0:9])
        iset2 = indexset.IndexSet(elements[10:19])

        self.assertTrue(iset1.isdisjoint(iset2))

        iset3 = iset1 | iset2
        self.assertTrue(iset3.issuperset(iset1))
        self.assertTrue(iset2.issubset(iset3))

        iset1 |= iset2
        self.assertTrue(iset1.issuperset(iset2))
        self.assertTrue(iset2.issubset(iset2))
        self.assertEqual(iset1, iset3)

    def testSetIntersection(self):
        elements = [FakeIndexable([i, "s%d" % i, (i, None)], i)
                    for i in xrange(20)]

        iset1 = indexset.IndexSet(elements[0:15])
        iset2 = indexset.IndexSet(elements[10:19])

        iset3 = iset1 & iset2
        self.assertEqual(len(iset3), 5)

        iset1 &= iset2
        self.assertItemsEqual(iset1, iset3)
        self.assertTrue(iset1 == iset3)



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER rule-based query analyzer.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

from efilter import engine
from efilter import errors
from efilter import protocol

from efilter.protocols import associative
from efilter.protocols import boolean
from efilter.protocols import name_delegate


class TypeInferencer(engine.VisitorEngine):
    """Determines the types of each subexpression and validates sanity.

    This class follows the visitor pattern. See documentation on VisitorEngine.
    """

    def visit_Literal(self, expr, **_):
        return type(expr.value)

    def visit_BinaryExpression(self, expr, **kwargs):
        lhst = self.visit(expr.lhs, **kwargs)
        if not protocol.isa(lhst, expr.type_signature[0]):
            raise errors.EfilterTypeError(query=self.query, root=expr.lhs,
                                          expected=expr.type_signature[0],
                                          actual=lhst)

        rhst = self.visit(expr.rhs, **kwargs)
        if not protocol.isa(rhst, expr.type_signature[1]):
            raise errors.EfilterTypeError(query=self.query, root=expr.rhs,
                                          expected=expr.type_signature[1],
                                          actual=rhst)

        return expr.return_signature

    def visit_Binding(self, expr, scope=None, **_):
        if isinstance(self.application_delegate, name_delegate.INameDelegate):
            result = name_delegate.reflect(self.application_delegate,
                                           expr.value,
                                           scope)
            if result:
                return result

        return protocol.AnyType

    def visit_VariadicExpression(self, expr, **kwargs):
        for subexpr in expr.children:
            t = self.visit(subexpr, **kwargs)
            if not protocol.isa(t, expr.type_signature):
                raise errors.EfilterTypeError(query=self.query,
                                              root=subexpr,
                                              expected=expr.type_signature,
                                              actual=t)

        return expr.return_signature

    def visit_Let(self, expr, scope=None, **kwargs):
        t = self.visit(expr.context, scope=scope, **kwargs)
        if not (t is protocol.AnyType
                or protocol.isa(t, associative.IAssociative)):
            raise errors.EfilterTypeError(query=self.query, root=expr,
                                          actual=t,
                                          expected=associative.IAssociative)

        return self.visit(expr.expression, scope=t, **kwargs)

    def visit_LetAny(self, expr, **kwargs):
        t = self.visit_Let(expr, **kwargs)
        if not protocol.isa(t, boolean.IBoolean):
            raise errors.EfilterTypeError(query=self.query, root=expr,
                                          actual=t,
                                          expected=associative.IBoolean)

        return boolean.IBoolean

    def visitLetEach(self, expr, **kwargs):
        return self.visit_LetAny(expr, **kwargs)


engine.Engine.register_engine(TypeInferencer, "infer_types")



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Canonical EFILTER syntax parser.

This is a syntax experiment, unifying destructuring operators using the
associative protocol. Basically, the slash and the arrow both become the dot.

It'll probably become the canonical syntax if it works well in Rekall.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"


import collections
import re

from efilter import errors
from efilter import expression
from efilter import frontend


# Transformation functions, for expressions that don't directly map to
# something in the AST.

def ComplementEquivalence(*args, **kwargs):
    return expression.Complement(
        expression.Equivalence(*args, **kwargs), **kwargs)


def ComplementMembership(*args, **kwargs):
    return expression.Complement(
        expression.Membership(*args, **kwargs), **kwargs)


def ReverseStrictOrderedSet(*args, **kwargs):
    return expression.StrictOrderedSet(*reversed(args), **kwargs)


def ReversePartialOrderedSet(*args, **kwargs):
    return expression.PartialOrderedSet(*reversed(args), **kwargs)


def NegateValue(*args, **kwargs):
    return expression.Product(
        expression.Literal(-1),
        *args,
        **kwargs)


def FlattenComponentLiteral(*args, **kwargs):
    if not isinstance(args[0], expression.Binding):
        raise ValueError(
            "'has component' must be followed by a component. Got %s." % (
                args[0]))
    return expression.ComponentLiteral(args[0].value, **kwargs)


def TransformLetAny(let, **kwargs):
    if not isinstance(let, expression.Let):
        raise ValueError("'any' must be followed by a 'where' expression.")
    context, expr = let.children
    return expression.LetAny(context, expr, **kwargs)


def TransformLetEach(let, **kwargs):
    if not isinstance(let, expression.Let):
        raise ValueError("'each' must be followed by a 'where' expression.")
    context, expr = let.children
    return expression.LetEach(context, expr, **kwargs)


# Operators - infix and prefix.

Operator = collections.namedtuple("Operator",
                                  "precedence assoc handler docstring")

# The order of precedence matters for generated matching rules, which is why
# this is an OrderedDict.
INFIX = collections.OrderedDict([
    ("+", Operator(precedence=3, assoc="left", handler=expression.Sum,
                   docstring="Arithmetic addition.")),
    ("-", Operator(precedence=3, assoc="left", handler=expression.Difference,
                   docstring="Arithmetic subtraction.")),
    ("*", Operator(precedence=5, assoc="left", handler=expression.Product,
                   docstring="Arithmetic multiplication.")),
    ("/", Operator(precedence=5, assoc="left", handler=expression.Quotient,
                   docstring="Arithmetic division.")),
    ("==", Operator(precedence=2, assoc="left",
                    handler=expression.Equivalence,
                    docstring="Equivalence (same as 'is').")),
    ("!=", Operator(precedence=2, assoc="left",
                    handler=ComplementEquivalence,
                    docstring="Inequivalence (same as 'is not').")),
    ("not in", Operator(precedence=2, assoc="left",
                        handler=ComplementMembership,
                        docstring="Left-hand operand is not in list.")),
    ("in", Operator(precedence=2, assoc="left",
                    handler=expression.Membership,
                    docstring="Left-hand operand is in list.")),
    (">", Operator(precedence=2, assoc="left",
                   handler=expression.StrictOrderedSet,
                   docstring="Greater-than.")),
    (">=", Operator(precedence=2, assoc="left",
                    handler=expression.PartialOrderedSet,
                    docstring="Equal-or-greater-than.")),
    ("<", Operator(precedence=2, assoc="left",
                   handler=ReverseStrictOrderedSet,
                   docstring="Less-than.")),
    ("<=", Operator(precedence=2, assoc="left",
                    handler=ReversePartialOrderedSet,
                    docstring="Equal-or-less-than.")),
    ("where", Operator(precedence=2, assoc="left",
                       handler=expression.Let,
                       docstring="VALUE where SUBEXPRESSION")),
    (".", Operator(precedence=6, assoc="left",
                   handler=expression.Let,
                   docstring="Shorthand for V where E.")),
    ("and", Operator(precedence=1, assoc="left",
                     handler=expression.Intersection,
                     docstring="Logical AND.")),
    ("or", Operator(precedence=0, assoc="left", handler=expression.Union,
                    docstring="Logical OR.")),
    ("=~", Operator(precedence=2, assoc="left",
                    handler=expression.RegexFilter,
                    docstring="Left-hand operand where regex.")),
])


PREFIX = {
    "not": Operator(precedence=5, assoc=None, handler=expression.Complement,
                    docstring="Logical NOT."),
    "-": Operator(precedence=4, assoc=None, handler=NegateValue,
                  docstring="Unary -."),
    "has component": Operator(precedence=7, assoc=None,
                              handler=FlattenComponentLiteral,
                              docstring="Matching entity must have component."),
    "any": Operator(precedence=1, assoc=None, handler=TransformLetAny,
                    docstring=("Following 'where' should succeed if "
                               "any left-hand value where.")),
    "each": Operator(precedence=1, assoc=None, handler=TransformLetEach,
                     docstring=("Following 'where' should only "
                                "succeed if all left-hand values match."))
}


def EnumAsPattern(strings):
    """Return a regex that'll match any of the strings in order."""
    return "(%s)" % "|".join([re.escape(x) for x in strings])


class Token(object):
    """Represents a result from the tokenizer."""

    def __init__(self, name, value, start, end):
        self.name = name
        self.value = value
        self.start = start
        self.end = end

    def __repr__(self):
        return "Token(name='%s', value='%s', start=%d, end=%d)" % (
            self.name, self.value, self.start, self.end)


class Pattern(object):
    """A token pattern.

    Args:
      state_regex: If this regular expression where the current state this
                   rule is considered.
      regex: A regular expression to try and match from the current point.
      actions: A command separated list of method names in the Lexer to call.
      next_state: The next state we transition to if this Pattern where.
      flags: flags to re.compile.
    """

    def __init__(self, label, state_regex, regex, actions, next_state,
                 flags=re.I):
        self.state_regex = re.compile(
            state_regex, re.DOTALL | re.M | re.S | re.U | flags)
        self.regex = re.compile(regex, re.DOTALL | re.M | re.S | re.U | flags)
        self.label = label
        self.re_str = regex

        if actions:
            self.actions = actions.split(",")
        else:
            self.actions = []

        self.next_state = next_state


class Tokenizer(object):
    """Context-free tokenizer for the efilter language.

    This is a very basic pattern-based tokenizer. Any rule from patterns
    will try to match the next token in the buffer if its state_regex where
    the current state. Only meaningful tokens are emitted (not whitespace.)
    """
    patterns = [
        # Keywords, operators and symbols
        Pattern("infix", "INITIAL", EnumAsPattern(INFIX.keys()),
                "emit", None),
        Pattern("prefix", "INITIAL", EnumAsPattern(PREFIX.keys()),
                "emit", None),
        Pattern("lparen", "INITIAL", r"\(",
                "emit", None),
        Pattern("rparen", "INITIAL", r"\)",
                "emit", None),
        Pattern("comma", "INITIAL", r",",
                "emit", None),
        Pattern("symbol", "INITIAL", r"([a-z_][a-z_0-9]+)", "emit", None),
        Pattern("param", "INITIAL", r"\{([a-z_0-9]*)\}", "emit_param", None),

        # Numeric literals
        Pattern("literal", "INITIAL", r"(\d+\.\d+)", "emit_float", None),
        Pattern("literal", "INITIAL", r"(0x[0-9a-zA-Z]+)", "emit_int16", None),
        Pattern("literal", "INITIAL", r"(\d+)", "emit_int", None),

        # String literals
        Pattern(None, "INITIAL", r"\"", "string_start", "STRING"),
        Pattern(None, "INITIAL", r"'", "string_start", "SQ_STRING"),

        Pattern("literal", "STRING", "\"", "pop_state,emit_string", None),
        Pattern(None, "STRING", r"\\(.)", "string_escape", None),
        Pattern(None, "STRING", r"[^\\\"]+", "string_append", None),

        Pattern("literal", "SQ_STRING", "'", "pop_state,emit_string", None),
        Pattern(None, "SQ_STRING", r"\\(.)", "string_escape", None),
        Pattern(None, "SQ_STRING", r"[^\\']+", "string_append", None),

        # Whitespace is ignored.
        Pattern(None, ".", r"\s+", None, None),
    ]

    def __init__(self, query):
        self.buffer = query
        self.state_stack = ["INITIAL"]
        self.current_token = None
        self._position = 0
        self.limit = len(query)
        self.lookahead = []
        self._param_idx = 0

    @property
    def position(self):
        """Returns the logical position (unaffected by lookahead)."""
        if self.lookahead:
            return self.lookahead[0].start

        return self._position

    def pop_state(self, **_):
        try:
            self.state_stack.pop()
        except IndexError:
            self.error("Pop state called on an empty stack.", self.position)

    def next_token(self):
        """Returns the next logical token.

        Will trigger parsing if it has to.
        """
        if self.lookahead:
            self.current_token = self.lookahead.pop(0)
            return self.current_token

        self.current_token = self._parse_next_token()
        return self.current_token

    def _parse_next_token(self):
        """Will parse patterns until it gets to the next token or EOF."""
        while self._position < self.limit:
            token = self.next_pattern()
            if token:
                return token

        return None

    def peek(self, steps=1):
        """Look ahead, doesn't affect current_token and next_token."""
        while len(self.lookahead) < steps:
            token = self._parse_next_token()
            if token is None:
                return None

            self.lookahead.append(token)

        return self.lookahead[steps - 1]

    def parse(self):
        """Yield every token in turn."""
        while self._position < self.limit:
            token = self.next_token()
            if not token:
                return

            yield token

    def next_pattern(self):
        """Parses the next pattern by matching each in turn."""
        current_state = self.state_stack[-1]
        position = self._position
        for pattern in self.patterns:
            if not pattern.state_regex.match(current_state):
                continue

            m = pattern.regex.match(self.buffer, position)
            if not m:
                continue

            position = m.end()
            token = None

            if pattern.next_state:
                self.state_stack.append(pattern.next_state)

            for action in pattern.actions:
                callback = getattr(self, action, None)
                if callback is None:
                    raise RuntimeError(
                        "No method defined for pattern action %s!" % action)

                token = callback(string=m.group(0), match=m, pattern=pattern)

            self._position = position

            return token

        self.error("Don't know how to match next. Did you forget quotes?",
                   self.position)

    def error(self, message, start, end=None):
        """Print a nice error."""
        raise errors.EfilterParseError(
            query=self.buffer, start=start, end=end, message=message)

    def emit(self, string, match, pattern, **_):
        """Emits a token using the current pattern match and pattern label."""
        return Token(name=pattern.label, value=string, start=match.start(),
                     end=match.end())

    def emit_param(self, match, pattern, **_):
        param_name = match.group(1)
        if not param_name:
            param_name = self._param_idx
            self._param_idx += 1

        return Token(name=pattern.label, value=param_name, start=match.start(),
                     end=match.end())

    def emit_int(self, string, match, pattern, **_):
        return Token(name=pattern.label, value=int(string), start=match.start(),
                     end=match.end())

    def emit_int16(self, string, match, pattern, **_):
        return Token(name=pattern.label, value=int(string, 16),
                     start=match.start(), end=match.end())

    def emit_float(self, string, match, pattern, **_):
        return Token(name=pattern.label, value=float(string),
                     start=match.start(), end=match.end())

    # String parsing

    def string_start(self, match, **_):
        self.string = ""
        self.string_position = match.start()

    def string_escape(self, string, match, **_):
        if match.group(1) in "'\"rnbt":
            self.string += string.decode("string_escape")
        else:
            self.string += string

    def string_append(self, string="", **_):
        self.string += string

    def emit_string(self, pattern, match, **_):
        return Token(name=pattern.label, value=self.string,
                     start=self.string_position, end=match.end())


class Parser(frontend.Frontend):
    """Parses the efilter language into the query AST.

    This is a basic precedence-climbing parser with support for prefix
    operators and a few special cases for list literals and such.
    """

    def __init__(self, original, params=None):
        super(Parser, self).__init__(original)

        self.tokenizer = Tokenizer(self.original)

        if isinstance(params, list):
            self.params = {}
            for idx, val in enumerate(params):
                self.params[idx] = val
        else:
            self.params = params

    def _handle_expr(self, operator, *args, **kwargs):
        try:
            return operator.handler(*args, **kwargs)
        except ValueError as e:
            return self.error(e.message,
                              start_token=args[0])

    def _replace_param(self, token):
        param_name = token.value
        value = self.params.get(param_name, None)
        if value is None:
            return self.error("No value provided for param %s" % param_name,
                              token)

        return value

    def next_atom(self):
        token = self.tokenizer.next_token()

        if token is None:
            return self.error("Unexpected end of input.")

        if token.name == "infix":
            if token.value == "-":
                # As it turns out, minus signs can be prefix operators! Who
                # knew? Certainly not the tokenizer.
                token.name = "prefix"
            else:
                return self.error("Unexpected infix operator.", token)

        if token.name == "prefix":
            operator = PREFIX[token.value]
            lhs = self.next_atom()
            rhs = self.next_expression(lhs, operator.precedence)
            return self._handle_expr(operator, rhs, start=token.start,
                                     end=rhs.end)

        if token.name == "literal":
            return expression.Literal(token.value, start=token.start,
                                      end=token.end)

        if token.name == "param":
            return expression.Literal(self._replace_param(token),
                                      start=token.start, end=token.end)

        if token.name == "symbol":
            return expression.Binding(token.value, start=token.start,
                                      end=token.end)

        if token.name == "lparen":
            # Parentheses can denote subexpressions or lists. Lists have at
            # least one comma before rparen (just like Python).
            lhs = self.next_atom()
            expr = self.next_expression(lhs, 0)
            if self.tokenizer.current_token is None:
                return self.error("End of input before closing parenthesis.",
                                  token)

            if self.tokenizer.peek().name == "comma":
                # It's a list, not an expression. Build it out as a literal.
                if not isinstance(lhs, expression.Literal):
                    return self.error(
                        "Non-literal value in list.", lhs)

                self.tokenizer.next_token()
                vals = [lhs.value]

                while (self.tokenizer.current_token and
                       self.tokenizer.current_token.name == "comma"):
                    atom = self.next_atom()
                    if not isinstance(atom, expression.Literal):
                        return self.error(
                            "Non-literal value in list", atom)
                    vals.append(atom.value)
                    self.tokenizer.next_token()

                if (self.tokenizer.current_token is None or
                        self.tokenizer.current_token.name != "rparen"):
                    self.error("Lists must end with a closing paren.",
                               self.tokenizer.current_token)

                return expression.Literal(tuple(vals), start=token.start,
                                          end=self.tokenizer.position)

            elif self.tokenizer.peek().name != "rparen":
                # We got here because there's still some stuff left to parse
                # and the next token is not an rparen. That can mean that an
                # infix operator is missing or that the parens are unmatched.
                # Decide which is more likely and raise the appropriate error.
                lparens = 1
                rparens = 0
                lookahead = 2
                while self.tokenizer.peek(lookahead):
                    if self.tokenizer.peek(lookahead).name == "lparen":
                        lparens += 1
                    elif self.tokenizer.peek(lookahead).name == "rparen":
                        rparens += 1

                    lookahead += 1

                if lparens > rparens:
                    return self.error("Ummatched left parenthesis.", token)
                else:
                    next_token = self.tokenizer.peek()
                    return self.error(
                        "Was not expecting %s here." % next_token.value,
                        next_token)

            self.tokenizer.next_token()
            return expr

        return self.error("Cannot handle token %s." % token, token)

    def next_expression(self, lhs, min_precedence):
        # This loop will spin as long as:
        # 1: There is a next token.
        # 2: It is an infix operator.
        # 3: Its precedence is higher than min_precedence.
        while self.tokenizer.peek():
            token = self.tokenizer.peek()

            if token.name != "infix":
                break

            operator = INFIX[token.value]
            if operator.precedence < min_precedence:
                break

            # We're a match - consume the next token.
            self.tokenizer.next_token()

            rhs = self.next_atom()
            next_min_precedence = operator.precedence
            if operator.assoc == "LEFT":
                next_min_precedence += 1

            # Let's see if the next infix operator (if any) is of higher
            # precedence than we are.
            while (self.tokenizer.peek() and
                   self.tokenizer.peek().name == "infix"):
                next_token = self.tokenizer.peek()
                next_operator = INFIX[next_token.value]
                if next_operator.precedence < next_min_precedence:
                    break
                rhs = self.next_expression(rhs, next_operator.precedence)

            lhs = self._handle_expr(operator, lhs, rhs, start=lhs.start,
                                    end=rhs.end)

        return lhs

    def parse(self):
        result = self.next_expression(self.next_atom(), 0)
        # If we didn't consume the whole query then raise.
        if self.tokenizer.peek():
            token = self.tokenizer.peek()
            return self.error(
                "Unexpected %s '%s'. Were you looking for an operator?" %
                (token.name, token.value),
                token)

        return result

    @property
    def root(self):
        return self.parse()

    def error(self, message, start_token=None, end_token=None):
        start = self.tokenizer.position
        end = start + 20
        if start_token:
            start = start_token.start
            end = start_token.end

        if end_token:
            end = end_token.end

        raise errors.EfilterParseError(
            query=self.original, start=start, end=end, message=message,
            token=start_token)


frontend.Frontend.register_frontend(Parser, shorthand="dotty")



from efilter import testlib

from efilter.protocols import boolean
from efilter.protocols import number
from efilter.protocols import name_delegate


class FakeApplicationDelegate(object):
    NAMES = {
        "ProcessName": basestring,
        "ProcessPid": int}

    def reflect(self, name, scope=None):
        _ = scope
        return self.NAMES.get(name)

    def provide(self, name):
        pass

    def getnames(self):
        return self.NAMES.iterkeys()


name_delegate.INameDelegate.implement(
    for_type=FakeApplicationDelegate,
    implementations={
        name_delegate.reflect: FakeApplicationDelegate.reflect,
        name_delegate.provide: FakeApplicationDelegate.provide,
        name_delegate.getnames: FakeApplicationDelegate.getnames
    }
)


class TypeInferenceTest(testlib.EngineTestCase):
    def testDelegate(self):
        self.assertEngineResult(
            engine="infer_types",
            source="ProcessName",
            expected=basestring,
            assertion=self.assertIsa)

    def testBasic(self):
        self.assertEngineResult(
            engine="infer_types",
            source="ProcessName == 'init'",
            expected=boolean.IBoolean,
            assertion=self.assertIsa)

        self.assertEngineResult(
            engine="infer_types",
            source="Process/Name == 'init'",
            source_syntax="slashy",
            expected=boolean.IBoolean,
            assertion=self.assertIsa)

    def testBinary(self):
        self.assertEngineResult(
            engine="infer_types",
            source="'foo' in ('bar', 'foo')",
            expected=boolean.IBoolean,
            assertion=self.assertIsa)

    def testRecursive(self):
        self.assertEngineResult(
            engine="infer_types",
            source="ProcessParent where (ProcessName == 'init')",
            expected=boolean.IBoolean,
            assertion=self.assertIsa)

        self.assertEngineResult(
            engine="infer_types",
            source="any ProcessChildren where (ProcessName == 'init')",
            expected=boolean.IBoolean,
            assertion=self.assertIsa)

    def testNumbers(self):
        self.assertEngineResult(
            engine="infer_types",
            source="5 + 5",
            expected=number.INumber,
            assertion=self.assertIsa)

        self.assertEngineResult(
            engine="infer_types",
            source="10 * (1 - 4) / 5",
            expected=number.INumber,
            assertion=self.assertIsa)

    def testDescendExpression(self):
        self.assertEngineResult(
            engine="infer_types",
            source="ProcessParent where (ProcessPid + 10)",
            expected=number.INumber,
            app_delegate=FakeApplicationDelegate(),
            assertion=self.assertIsa)

        # Should be the same using shorthand syntax.
        self.assertEngineResult(
            engine="infer_types",
            source="ProcessParent.ProcessParent where (ProcessPid - 1)",
            expected=number.INumber,
            app_delegate=FakeApplicationDelegate(),
            assertion=self.assertIsa)



from efilter import testlib


class MatcherTest(testlib.EngineTestCase):
    def assertMatcherResult(self, expected, result):
        self.assertEqual(expected, result.result)

    def testBasic(self):
        self.assertEngineResult(
            engine="filter",
            assertion=self.assertMatcherResult,
            source="Process/pid is 1",
            source_syntax="slashy",
            expected=True,
            bindings={"Process/pid": 1})

    def testRecursion(self):
        # Drilling down into a nested object should succeed.
        self.assertEngineResult(
            engine="filter",
            assertion=self.assertMatcherResult,
            source="Process/parent matches Process/pid is 1",
            bindings={"Process/parent": {"Process/pid": 1}},
            expected=True,
            source_syntax="slashy")

        # Using a let-any form should succeed even if there is only one linked
        # object.
        self.assertEngineResult(
            engine="filter",
            assertion=self.assertMatcherResult,
            source=("any Process/parent matches (Process/pid is 1 or "
                    "Process/command is 'foo')"),
            bindings={"Process/parent": {"Process/pid": 1}},
            expected=True,
            source_syntax="slashy")



import unittest

from efilter import dispatch
from efilter import protocol


@dispatch.polymorphic
def say_moo(cow):
    _ = cow
    raise NotImplementedError()


@dispatch.polymorphic
def graze(cow):
    _ = cow
    raise NotImplementedError()


class IBovine(protocol.Protocol):
    _protocol_functions = (say_moo, graze)


class Kyr(object):
    def say_muu(self):
        return "Muu"


IBovine.implement(for_type=Kyr,
                  implementations={
                      graze: lambda c: "Om nom nom.",
                      say_moo: lambda c: c.say_muu()})


class TypesTest(unittest.TestCase):
    def testProtocol(self):
        self.assertTrue(isinstance(Kyr(), IBovine))
        self.assertEquals(say_moo(Kyr()), "Muu")
        self.assertEqual(graze(Kyr()), "Om nom nom.")



from efilter import testlib


class HinterTest(testlib.EngineTestCase):
    def testNop(self):
        self.assertTransform(
            engine="hinter",
            original="Process.name == 'init'",
            expected="Process.name == 'init'",
            selector=None)

    def testBasic(self):
        self.assertTransform(
            engine="hinter",
            original="Process.name == 'init'",
            expected="name == 'init'",
            selector="Process")

    def testMulti(self):
        self.assertTransform(
            engine="hinter",
            original="Process.parent.Process.name == 'init'",
            selector="Process.parent",
            expected="Process.name == 'init'")

    def testNested(self):
        self.assertTransform(
            engine="hinter",
            original=("Process.parent.Process.name == 'init' "
                      "and Process.pid > 10"),
            selector="Process.parent",
            expected="Process.name == 'init'")

    def testAndNested(self):
        self.assertTransform(
            engine="hinter",
            original=("Process.parent.Process.name == 'init' "
                      "and Process.parent.Process.pid > 10"),
            selector="Process.parent",
            expected="Process.name == 'init' and Process.pid > 10")

    def testNestedWithComplement(self):
        self.assertTransform(
            engine="hinter",
            original=("Process.parent.Process.name != 'init' "
                      "and not Process.parent.Process.pid > 10"),
            selector="Process.parent",
            expected="Process.name != 'init' and not Process.pid > 10")

    def testLegacy(self):
        self.assertTransform(
            engine="hinter",
            original=("MemoryDescriptor.process where "
                      "(Process.command == 'Adium') "
                      "and 'execute' in MemoryDescriptor.permissions "
                      "and 'write' in MemoryDescriptor.permissions"),
            selector="MemoryDescriptor.process",
            expected="Process.command == 'Adium'")



# -*- coding: utf-8 -*-

# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""EFILTER abstract type system."""

from efilter import dispatch
from efilter import protocol

from efilter.protocols import eq

# Declarations:
# pylint: disable=unused-argument


@dispatch.polymorphic
def union(x, y):
    raise NotImplementedError()


@dispatch.polymorphic
def intersection(x, y):
    raise NotImplementedError()


@dispatch.polymorphic
def difference(x, y):
    raise NotImplementedError()


@dispatch.polymorphic
def issuperset(x, y):
    raise NotImplementedError()


@dispatch.polymorphic
def issubset(x, y):
    return issuperset(y, x)


@dispatch.polymorphic
def isstrictsuperset(x, y):
    return issuperset(x, y) and eq.ne(x, y)


@dispatch.polymorphic
def isstrictsubset(x, y):
    return isstrictsuperset(y, x)


@dispatch.polymorphic
def contains(s, e):
    raise NotImplementedError()


class ISet(protocol.Protocol):
    _protocol_functions = (union, intersection, difference, issuperset,
                           contains)


# Default implementations:

ISet.implement(
    for_types=(set, frozenset),
    implementations={
        union: lambda x, y: x | frozenset(y),
        intersection: lambda x, y: x & frozenset(y),
        difference: lambda x, y: x - frozenset(y),
        issuperset: lambda x, y: x >= frozenset(y),
        contains: lambda s, e: e in s
    }
)


ISet.implement(
    for_types=(list, tuple),
    implementations={
        union: lambda x, y: frozenset(x) | frozenset(y),
        intersection: lambda x, y: frozenset(x) & frozenset(y),
        difference: lambda x, y: frozenset(x) - frozenset(y),
        issuperset: lambda x, y: frozenset(x) >= frozenset(y),
        contains: lambda s, e: e in s
    }
)



import unittest

from efilter.ext import superposition


class SuperpositionTest(unittest.TestCase):
    """This file only tests the specifities of ext.superpositions.

    efilter.protocols.test_superpositon has additional tests, based on the
    generic protocol.
    """

    def testMutability(self):
        """Test adding states."""
        s = superposition.HashedSuperposition(1, 2, 3)
        s.add_state(4)
        self.assertEqual(sorted(s.getstates()), [1, 2, 3, 4])

        # Adding another superposition should leave us flat.
        s.add_state(superposition.HashedSuperposition(4, 5))
        self.assertEqual(sorted(s.getstates()), [1, 2, 3, 4, 5])



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER query wrapper.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"


from efilter import frontend
from efilter import engine
from efilter import expression


class Query(object):
    source = None
    root = None
    syntax = None

    def __init__(self, source, root=None, params=None, syntax="slashy",
                 application_delegate=None):
        super(Query, self).__init__()
        self.application_delegate = application_delegate
        self.syntax = syntax

        # Run as a copy constructor.
        if isinstance(source, Query):
            self.source = source.source
            self.root = root or source.root
            self.application_delegate = (application_delegate
                                         or source.application_delegate)
            return

        # No need to parse anything.
        if root:
            self.source = source
            self.root = root
            return

        # This is here to support calling Query with a positional argument
        # that's the root of the AST.
        if isinstance(source, expression.Expression):
            self.root = source
            return

        # If we got here then source is actually a query and it's not been
        # parsed, so we need to find the appropriate frontend parser and run it.
        if self.syntax and self.syntax != "expression":
            # We need to parse the query.
            parser_cls = frontend.Frontend.get_frontend(syntax)
            parser = parser_cls(original=source, params=params)
            self.source = parser.original
            self.root = parser.root

            # Query is normalized by default because parsers are allowed to
            # produce crazy AST and rely on the visitor engines to understand
            # it.
            self.root = self.run_engine("normalizer").root
        else:
            raise TypeError("%r is not an expression.", source)

    def subquery(self, node):
        return type(self)(self, root=node)

    def locate_expression(self, node):
        """Returns the original source of the expression with context.

        Returns tuple of:
            - query up to the start of expression
            - source of the expression
            - the rest of the query
        """
        return node.start, node.end

    def source_expression(self, node):
        if isinstance(self.source, basestring):
            return self.source[node.start:node.end]

        return None

    def run_engine(self, shorthand, **kwargs):
        engine_cls = engine.Engine.get_engine(shorthand)
        if engine_cls is None:
            raise ValueError("No such engine %r." % shorthand)

        return engine_cls(
            query=self,
            application_delegate=self.application_delegate).run(**kwargs)

    def __str__(self):
        return unicode(self)

    def __unicode__(self):
        if self.source:
            return unicode(self.source)

        return unicode(self.root)

    def __repr__(self):
        source = self.source_expression(self.root)
        if not source:
            source = self.root

        return "Query(%r)" % source

    def __hash__(self):
        return hash(self.root)

    def __eq__(self, other):
        if not isinstance(other, Query):
            return False

        return self.root == other.root

    def __ne__(self, other):
        return not self.__eq__(other)



import unittest

from efilter.frontends.experiments import dotty
from efilter import expression


class TokenizerTest(unittest.TestCase):
    def assertQueryMatches(self, query, expected):
        tokenizer = dotty.Tokenizer(query)
        actual = [(token.name, token.value) for token in tokenizer.parse()]
        self.assertEqual(expected, actual)

    def testLiterals(self):
        queries = [
            ("0xf", [15]),
            ("234.7  15\n ", [234.7, 15]),
            ("  15 0x15 '0x15' ' 52.6'", [15, 21, "0x15", " 52.6"])]

        for query, values in queries:
            expected = [("literal", val) for val in values]
            self.assertQueryMatches(query, expected)

    def testKeywords(self):
        query = "5 + 5 == 10 and 'foo' =~ 'foo'"
        expected = [
            ("literal", 5),
            ("infix", "+"),
            ("literal", 5),
            ("infix", "=="),
            ("literal", 10),
            ("infix", "and"),
            ("literal", "foo"),
            ("infix", "=~"),
            ("literal", "foo")]
        self.assertQueryMatches(query, expected)

    def testWhitespace(self):
        query = "20 not in 10"
        expected = [
            ("literal", 20),
            ("infix", "not in"),
            ("literal", 10)]
        self.assertQueryMatches(query, expected)

    def testLists(self):
        query = "'foo' in ('foo', 'bar')"
        expected = [
            ("literal", "foo"),
            ("infix", "in"),
            ("lparen", "("),
            ("literal", "foo"),
            ("comma", ","),
            ("literal", "bar"),
            ("rparen", ")")]
        self.assertQueryMatches(query, expected)

    def testPeeking(self):
        query = "1 in (5, 10) == ProcessPid"
        tokenizer = dotty.Tokenizer(query)
        tokenizer.next_token()
        self.assertEquals(tokenizer.peek(2).name, "lparen")
        self.assertEquals(tokenizer.current_token.value, 1)
        self.assertEquals(tokenizer.peek(20), None)
        self.assertEquals(tokenizer.current_token.value, 1)
        self.assertEquals(tokenizer.next_token().value, "in")
        self.assertEquals(tokenizer.current_token.value, "in")
        self.assertEquals(tokenizer.next_token().name, "lparen")
        self.assertEquals(tokenizer.next_token().value, 5)
        self.assertEquals(tokenizer.peek().name, "comma")
        self.assertEquals(tokenizer.next_token().name, "comma")
        self.assertEquals(tokenizer.next_token().value, 10)


class ParserTest(unittest.TestCase):
    def assertQueryMatches(self, query, expected, params=None):
        parser = dotty.Parser(query, params=params)
        actual = parser.parse()
        self.assertEqual(expected, actual)

    def assertQueryRaises(self, query, params=None):
        parser = dotty.Parser(query, params=params)
        self.assertRaises(dotty.errors.EfilterParseError, parser.parse)

    def testLiterals(self):
        query = "0xff"
        expected = expression.Literal(255)
        self.assertQueryMatches(query, expected)

    def testDescendQuery(self):
        query = "Process where (name == 'init' and pid == 1)"
        expected = expression.Let(
            expression.Binding("Process"),
            expression.Intersection(
                expression.Equivalence(
                    expression.Binding("name"),
                    expression.Literal("init")
                ),
                expression.Equivalence(
                    expression.Binding("pid"),
                    expression.Literal(1))))
        self.assertQueryMatches(query, expected)

    def testDescendShorthand(self):
        query = "ProcessParent.ProcessParent.ProcessName"
        expected = expression.Let(
            expression.Binding("ProcessParent"),
            expression.Let(
                expression.Binding("ProcessParent"),
                expression.Binding("ProcessName")))
        self.assertQueryMatches(query, expected)

    def testEquivalence(self):
        query = "10 == 10"
        expected = expression.Equivalence(
            expression.Literal(10),
            expression.Literal(10))
        self.assertQueryMatches(query, expected)

    def testPrecedence(self):
        query = "5 == 1 * 5 and ProcessName == 'init'"
        expected = expression.Intersection(
            expression.Equivalence(
                expression.Literal(5),
                expression.Product(
                    expression.Literal(1),
                    expression.Literal(5))),
            expression.Equivalence(
                expression.Binding("ProcessName"),
                expression.Literal("init")))
        self.assertQueryMatches(query, expected)

    def testParensBaseline(self):
        query = "3 + 2 * 5"
        expected = expression.Sum(
            expression.Literal(3),
            expression.Product(
                expression.Literal(2),
                expression.Literal(5)))

        self.assertQueryMatches(query, expected)

    def testParens(self):
        query = "(3 + 2) * 5"
        expected = expression.Product(
            expression.Sum(
                expression.Literal(3),
                expression.Literal(2)),
            expression.Literal(5))

        self.assertQueryMatches(query, expected)

    def testPrefixMinus(self):
        query = "-(5 + 5)"
        expected = expression.Product(
            expression.Literal(-1),
            expression.Sum(
                expression.Literal(5),
                expression.Literal(5)))

        self.assertQueryMatches(query, expected)

    def testPrefixMinusHighPrecedence(self):
        query = "-5 + 5"
        expected = expression.Sum(
            expression.Product(
                expression.Literal(-1),
                expression.Literal(5)),
            expression.Literal(5))

        self.assertQueryMatches(query, expected)

    def testPrefixMinusLowPrecedence(self):
        query = "-5 * 5"
        expected = expression.Product(
            expression.Literal(-1),
            expression.Product(
                expression.Literal(5),
                expression.Literal(5)))

        self.assertQueryMatches(query, expected)

    def testLetSingle(self):
        query = "ProcessParent where ProcessCommand == 'init'"
        expected = expression.Let(
            expression.Binding("ProcessParent"),
            expression.Equivalence(
                expression.Binding("ProcessCommand"),
                expression.Literal("init")))

        self.assertQueryMatches(query, expected)

    def testLetSubexpr(self):
        query = ("ProcessParent where (ProcessCommand == 'init' and "
                 "ProcessPid == 1)")
        expected = expression.Let(
            expression.Binding("ProcessParent"),
            expression.Intersection(
                expression.Equivalence(
                    expression.Binding("ProcessCommand"),
                    expression.Literal("init")),
                expression.Equivalence(
                    expression.Binding("ProcessPid"),
                    expression.Literal(1))))

        self.assertQueryMatches(query, expected)

    def testLetSingleAny(self):
        query = "any Process.parent where name == 'init'"
        expected = expression.LetAny(
            expression.Let(
                expression.Binding("Process"),
                expression.Binding("parent")),
            expression.Equivalence(
                expression.Binding("name"),
                expression.Literal("init")))

        self.assertQueryMatches(query, expected)

    def testLetSubexprEach(self):
        query = "each ProcessChildren where ProcessCommand == 'foo'"
        expected = expression.LetEach(
            expression.Binding("ProcessChildren"),
            expression.Equivalence(
                expression.Binding("ProcessCommand"),
                expression.Literal("foo")))

        self.assertQueryMatches(query, expected)

    def testLists(self):
        query = "'foo' in ('foo', 'bar') and 1 not in (5, 2, 3,17)"
        expected = expression.Intersection(
            expression.Membership(
                expression.Literal("foo"),
                expression.Literal(("foo", "bar"))),
            expression.Complement(
                expression.Membership(
                    expression.Literal(1),
                    expression.Literal((5, 2, 3, 17)))))

        self.assertQueryMatches(query, expected)

    def testBigQuery(self):
        query = ("(ProcessPid == 1 and ProcessCommand in ('init', 'initd')) "
                 "or any ProcessChildren where (ProcessCommand not in "
                 "('launchd', 'foo'))")
        expected = expression.Union(
            expression.Intersection(
                expression.Equivalence(
                    expression.Binding("ProcessPid"),
                    expression.Literal(1)),
                expression.Membership(
                    expression.Binding("ProcessCommand"),
                    expression.Literal(("init", "initd")))),
            expression.LetAny(
                expression.Binding("ProcessChildren"),
                expression.Complement(
                    expression.Membership(
                        expression.Binding("ProcessCommand"),
                        expression.Literal(("launchd", "foo"))))))

        self.assertQueryMatches(query, expected)

    def testLooseAnyError(self):
        query = "any ProcessCommand == 'init'"
        self.assertQueryRaises(query)

    def testMissingClosingParens(self):
        query = "ProcessPid in (1,5"
        self.assertQueryRaises(query)

    def testNestedParens(self):
        query = "ProcessPid in ((1,2))"
        expected = expression.Membership(
            expression.Binding("ProcessPid"),
            expression.Literal((1, 2)))
        self.assertQueryMatches(query, expected)

    def testHasComponent(self):
        query = "has component Process"
        expected = expression.ComponentLiteral("Process")
        self.assertQueryMatches(query, expected)

    def testTemplateReplacements(self):
        query = "ProcessPid == {}"
        params = [1]
        exptected = expression.Equivalence(
            expression.Binding("ProcessPid"),
            expression.Literal(1))
        self.assertQueryMatches(query, exptected, params=params)

        query = "ProcessPid == {pid}"
        params = {"pid": 1}
        exptected = expression.Equivalence(
            expression.Binding("ProcessPid"),
            expression.Literal(1))
        self.assertQueryMatches(query, exptected, params=params)

    def testParamFailures(self):
        query = "{foo} == 1"
        params = ["ProcessPid"]
        self.assertQueryRaises(query, params=params)

        # Even fixing the above, the left side should be a literal, not a
        # binding.
        query = "{foo} == 1"
        params = {"foo": "ProcessPid"}
        exptected = expression.Equivalence(
            expression.Literal("ProcessPid"),
            expression.Literal(1))
        self.assertQueryMatches(query, exptected, params=params)

    def testParenParsing(self):
        # This query should fail on the lose 'name' token:
        query = ("BufferPurpose == 'zones' and any BufferContext where"
                 " (AllocationZone name == {zone_name})")
        params = dict(zone_name="foo")
        parser = dotty.Parser(query, params=params)
        try:
            parser.parse()
        except dotty.errors.EfilterParseError as e:
            self.assertEqual(e.token.value, 'name')

    def testMultipleLiterals(self):
        query = "ProcessBinding foo foo bar 15"
        self.assertQueryRaises(query)



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER special frontends.

Frontends in this module don't really implement a language - they're special
cases for just passing through literals and stuff.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

from efilter import expression
from efilter import frontend


class LiteralFrontend(frontend.Frontend):
    """This is basically and identity function for literals."""

    @property
    def root(self):
        return expression.Literal(self.original)


frontend.Frontend.register_frontend(LiteralFrontend, shorthand="literal")


class PassthroughFrontend(frontend.Frontend):
    """This is basically and identity function for expressions."""

    @property
    def root(self):
        return self.original


frontend.Frontend.register_frontend(LiteralFrontend, shorthand="expression")



# -*- coding: utf-8 -*-

# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""EFILTER abstract type system.

The superposition protocol concerns itself with variables that may have several
possible values, and it is unclear which value is the correct one, or whether
there even is a correct value.

Real-world examples of when this happens include merge conflicts, roots of
quadratic equations, and programmers suffering from pathological inability to
make decisions.
"""

from efilter import dispatch
from efilter import protocol

from efilter.protocols import eq

# Declarations:
# pylint: disable=unused-argument


@dispatch.polymorphic
def superposition(first_state, *states):
    """Build a superposition of states, all of which must be the same type."""
    raise NotImplementedError()


def meld(*states):
    """Returns a superposition of states, or the state if all args are equal.

    This is a convenience function, equivalent to calling
    getstate(superposition(x)) to get x.

    This function skips over instances of None in states (None is not allowed)
    in superpositions.

    Examples:
        meld("foo", "bar") # => HashedSuperposition("foo", "bar")
        meld("foo", "foo") # => "foo"
        meld("foo", None) # => "foo"
        meld(None) # => None
    """
    states = [x for x in states if x is not None]
    if not states:
        return None

    s = superposition(*states)
    if insuperposition(s):
        return s

    return getstate(s)


@dispatch.polymorphic
def getstates(x):
    """Return a collection of the possible states of x."""
    raise NotImplementedError()


def getstate(x):
    """Return the state of x, or raise exception if x is in superposition."""
    if insuperposition(x):
        raise TypeError(
            "Ambiguous call to getstate for %r which has more than one state."
            % x)

    for state in getstates(x):
        return state


@dispatch.polymorphic
def state_type(x):
    """Return the type (class) of the states of x."""
    raise NotImplementedError()


@dispatch.polymorphic
def state_union(x, y):
    """Return a new superposition with the union of states."""
    raise NotImplementedError()


@dispatch.polymorphic
def state_intersection(x, y):
    """Return a new superposition with the intersection of states."""
    raise NotImplementedError()


@dispatch.polymorphic
def state_difference(x, y):
    """Return a new superposition with the difference of states of x and y."""
    raise NotImplementedError()


@dispatch.polymorphic
def hasstate(sp, state):
    """Does superposition have the (scalar) state?

    Calling this with a state that is, itself, a superposition must return
    False. (Use state_superset for comparing two superpositions in a similar
    fashion.)
    """
    raise NotImplementedError()


@dispatch.polymorphic
def state_eq(x, y):
    """Are the states of x and y exactly the same?"""
    raise NotImplementedError()


@dispatch.polymorphic
def state_superset(x, y):
    """Are the states of x a superset of the states of y?"""
    raise NotImplementedError()


@dispatch.polymorphic
def state_subset(x, y):
    """Optional: are the states of x a subset of the states of y?"""
    return state_superset(y, x)


@dispatch.polymorphic
def state_strictsuperset(x, y):
    """Optional: Are the states of x a strict superset of the states of y?"""
    return state_superset(x, y) and not state_eq(x, y)


@dispatch.polymorphic
def state_strictsubset(x, y):
    """Optional: Are the states of x a strict subset of the states of y?"""
    return state_strictsuperset(y, x)


@dispatch.polymorphic
def state_apply(x, f):
    """Apply f to each state of x and return a new superposition of results."""
    raise NotImplementedError()


@dispatch.polymorphic
def insuperposition(x):
    """Optional: Is x a superposition AND does it have more than one state?"""
    return isinstance(x, ISuperposition) and len(getstates(x)) > 1


class ISuperposition(protocol.Protocol):
    _protocol_functions = (getstates, state_type, state_union,
                           state_intersection, state_difference, hasstate,
                           state_eq, state_superset, state_apply)


def _scalar_hasstate(sp, state):
    if insuperposition(state):
        raise TypeError(
            "2nd argument to hasstate must not be in superposition. Got %r."
            % state)

    return state_eq(sp, state)


def _scalar_state_eq(x, y):
    if insuperposition(y):
        return False

    return eq.eq(x, getstate(y))


# Implementation for scalars:
# pylint: disable=unnecessary-lambda
ISuperposition.implement(
    for_type=protocol.AnyType,
    implementations={
        getstates: lambda x: (x,),
        state_type: lambda x: type(x),
        state_union: lambda x, y: superposition(x, y),
        state_intersection: lambda x, y: x if hasstate(y, x) else None,
        state_difference: lambda x, y: None if hasstate(y, x) else x,
        hasstate: _scalar_hasstate,
        state_eq: _scalar_state_eq,
        state_superset: lambda x, y: state_eq(x, y),
        state_apply: lambda x, f: f(x)
    }
)



# -*- coding: utf-8 -*-

# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER abstract type system.

The type protocols defined under efilter.protocols.* provide a very thin layer
over Python's builtin types, defined as collections of related functions with
defined semantics. Each type protocol is intended to uniformly support a
specific behavior across any type that participates in the protocol.

To participate in a protocol, two things are required:
1) Implementations of each of the member functions must be provided.
2) The type must be formally added to the protocol.

In this manner, we are able to declare strict compositional types on atoms and
expressions in the EFILTER AST and allow type hierarchies external to EFILTER
(Plaso Events, Rekall Entities) to be passed to the EFILTER engines without
casting or wrapping.

The compositional, flat nature of the type protocols makes it simple to support
basic type inference, by annotating each expression type with sets of
protocols it requires on its children and guarantees on its return type.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

import abc


class AnyType(object):
    """Sentinel used to provide a default implementation of a protocol.

    If you need to provide a default implementation of functions in a
    protocol (for example, providing fall-through behavior for objects that
    don't participate in the protocol) you may pass this type in place of
    'object'. This will cause the polymorphic functions to fall through to
    this default implementation, but won't cause 'object' to be a subclass
    of the protocol.

    Example:
        MyProtocol.implement(for_type=AnyType,
                             implementations={foo=lambda x: "foo"})

        foo(5)  # => "foo"
        isinstance(5, MyProtocol)  # => False
        implements(5, MyProtocol)  # => True
    """


BUILTIN_TYPES = (int, float, long, complex, basestring, tuple, list, dict, set,
                 frozenset, type(None), AnyType)


def implements(obj, protocol):
    """Does the object 'obj' implement the 'prococol'?"""
    if isinstance(obj, type):
        raise TypeError("First argument to implements must be an instance. "
                        "Got %r." % obj)
    return isinstance(obj, protocol) or issubclass(AnyType, protocol)


def isa(cls, protocol):
    """Does the type 'cls' participate in the 'protocol'?"""
    if not isinstance(cls, type):
        raise TypeError("First argument to isa must be a type. Got %r." % cls)
    return issubclass(cls, protocol) or issubclass(AnyType, protocol)


class Protocol(object):
    """Collection of related functions that operate on a type (interface)."""
    __metaclass__ = abc.ABCMeta

    _protocol_functions = set()

    @classmethod
    def functions(cls):
        result = set(cls._protocol_functions)

        for scls in cls.mro():
            protocol_functions = getattr(scls, "_protocol_functions", None)
            if protocol_functions:
                result.update(protocol_functions)

        return result

    @classmethod
    def implemented(cls, for_type):
        for function in cls.functions():
            if not function.implemented_for_type(for_type):
                raise TypeError(
                    "%r doesn't implement %r so it cannot participate in "
                    "the protocol %r." %
                    (for_type, function.func.func_name, cls))

        cls.register(for_type)

    @classmethod
    def _implement_for_type(cls, for_type, implementations):
        # AnyType is a sentinel that means the polymorphic function should
        # just dispatch on 'object'.
        dispatch_type = object if for_type is AnyType else for_type
        protocol_functions = cls.functions()
        remaining = set(protocol_functions)

        for func, impl in implementations.iteritems():
            if func not in protocol_functions:
                func_name = getattr(func, "func_name", repr(func))
                raise TypeError("Function %s is not part of the protocol %r." %
                                (func_name, cls))

            func.implement(for_type=dispatch_type,
                           implementation=impl)
            remaining.remove(func)

        if remaining:
            raise TypeError(
                "%s.implement invokation must provide implementations of "
                "%r" % (cls.__name__, remaining))

        cls.implemented(for_type=for_type)

    @classmethod
    def implement(cls, implementations, for_type=None, for_types=None):
        """Provide protocol implementation for a type.

        Register all implementations of polymorphic functions in this
        protocol and adds the type into the abstract base class of the
        protocol.

        Arguments:
            implementations: A dict of (function, implementation), where each
                function is polymorphic and each implementation is a callable.
            for_type: The concrete type implementations apply to.
            for_types: Same as for_type, but takes a tuple of types.

            You may not supply both for_type and for_types for obvious reasons.

        Raises:
            ValueError for arguments.
            TypeError if not all implementations are provided or if there
                are issues related to polymorphism (e.g. attempting to
                implement a non-polymorphic function.
        """
        if for_type:
            if for_types:
                raise ValueError("Cannot pass both for_type and for_types.")
            for_types = (for_type,)
        elif for_types:
            if not isinstance(for_types, tuple):
                raise TypeError("for_types must be passed as a tuple of "
                                "types (classes).")
        else:
            raise ValueError("Must pass either for_type or for_types.")

        for type_ in for_types:
            cls._implement_for_type(for_type=type_,
                                    implementations=implementations)



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER rule-based query analyzer.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

import collections

from efilter import engine
from efilter import errors
from efilter import expression
from efilter import protocol

from efilter.protocols import iset

Analysis = collections.namedtuple("Analysis",
                                  ("symbols", "eq_indexables"))


class RuleAnalyzer(engine.VisitorEngine):
    """This is a rule-driven analyzer that gets a list of symbols and indexing.

    This class follows the visitor pattern. See documentation on VisitorEngine.

    The analyzer will produce a list of symbols required by the query (based on
    the Bindings/variables) and recommend a list of Bindings suitable for
    building an equivalence-based index (based on Equivalence expressions in
    the query).
    """

    def visit_Expression(self, expr, **_):
        _ = expr
        return Analysis((), ())

    def visit_Binding(self, expr, **_):
        return Analysis((expr.value,), ())

    def visit_BinaryExpression(self, expr, **kwargs):
        lhsa = self.visit(expr.lhs, **kwargs)
        rhsa = self.visit(expr.rhs, **kwargs)

        return Analysis(
            iset.union(lhsa.symbols, rhsa.symbols),
            iset.union(lhsa.eq_indexables, rhsa.eq_indexables))

    def visit_VariadicExpression(self, expr, **kwargs):
        symbols = set()
        eq_indexables = set()

        for child in expr:
            analysis = self.visit(child, **kwargs)
            symbols.update(analysis.symbols)
            eq_indexables.update(analysis.eq_indexables)

        return Analysis(symbols, eq_indexables)

    def visit_Let(self, expr, scope=None, **kwargs):
        if not isinstance(expr.lhs, expression.Binding):
            # Technically, the LHS context can be anything that implements
            # IAssociative, so a literal, or a subexpression that evaluates to
            # one are possible. Unfortunately, when that happens it is
            # non-trivial (read hard (read impossible)) to correctly determine
            # the scope for the RHS of the Let-form.
            #
            # As this is the case, we are unable to create any hints, and
            # any symbols in the RHS expression are bound to an anonymous scope
            # and, and, as such, not useful.
            return self.visit_BinaryExpression(expr, scope=scope, **kwargs)

        scope = expr.lhs.value
        rhsa = self.visit(expr.rhs, scope=scope, **kwargs)
        symbols = set("%s.%s" % (scope, symbol) for symbol in rhsa.symbols)
        symbols.update(rhsa.symbols)
        symbols.add(expr.lhs.value)
        return rhsa._replace(symbols=symbols)

    def visit_Membership(self, expr, **kwargs):
        symbols = set()
        lha = self.visit(expr.lhs, **kwargs)
        rha = self.visit(expr.rhs, **kwargs)
        symbols.update(lha.symbols)
        symbols.update(rha.symbols)

        if (not isinstance(expr.rhs, expression.Literal)
                or not isinstance(expr.lhs, expression.Binding)):
            return Analysis(symbols, ())

        if not protocol.implements(expr.rhs.value, iset.ISet):
            # Yup, no can do.
            raise errors.EfilterTypeError(root=expr.rhs, query=self.query,
                                          actual=type(expr.rhs.value),
                                          expected=iset.ISet)

        return Analysis(symbols, (expr.lhs.value,))

    def visit_Equivalence(self, expr, **kwargs):
        literal = None
        indexables = set()
        symbols = set()
        for child in expr.children:
            if isinstance(child, expression.Literal):
                if literal is not None and literal != child.value:
                    # This means something like 5 == 6 is being asked. This
                    # expression will always be false and it makes no sense to
                    # continue.
                    return Analysis((), ())
                else:
                    literal = child.value
            elif isinstance(child, expression.Binding):
                indexables.add(child.value)
                symbols.add(child.value)
            elif isinstance(child, expression.Let):
                # If we get a let-form, follow down as long as RHS is another
                # left form and the LHS is a binding. (something like
                # foo.bar.baz)
                let = child
                path = []
                while (isinstance(let, expression.Let)
                       and isinstance(let.lhs, expression.Binding)):
                    path.append(let.lhs.value)
                    symbols.add(".".join(path))
                    let = let.rhs

                if isinstance(let, expression.Binding):
                    path.append(let.value)

                remainder = self.visit(child, **kwargs)
                symbols.update(remainder.symbols)

                symbol = ".".join(path)
                symbols.add(symbol)
                indexables.add(symbol)
            else:
                analysis = self.visit(child, **kwargs)
                symbols.update(analysis.symbols)

        return Analysis(symbols, indexables)


engine.Engine.register_engine(RuleAnalyzer, "analyzer")



"""EFILTER Forensic Query Language"""

from efilter.engines import analyzer
from efilter.engines import dotty_output
from efilter.engines import hinter
from efilter.engines import matcher
from efilter.engines import normalizer
from efilter.engines import type_inferencer



from efilter import expression
from efilter import testlib


class NormalizerTest(testlib.EngineTestCase):
    def testBasic(self):
        self.assertTransform("normalizer", "ProcessName == 'init'",
                             "ProcessName == 'init'")

    def testEliminateEmpties(self):
        self.assertTransform(
            "normalizer", ("&", True), expression.Literal(True))

    def testLetForms(self):
        """LetAny and LetEach forms are not rotated."""
        original = ("let-any",
                    ("let", ("var", "Process"), ("var", "parent")),
                    ("==", ("var", "name"), "init"))

        self.assertTransform("normalizer", original, original)

    def testRealExample(self):
        original = ("&",
                    ("let",
                     ("let", ("var", "MemoryDescriptor"), ("var", "process")),
                     ("==",
                      ("let", ("var", "Process"), ("var", "command")),
                      "Adium")),
                    ("&",
                     ("in", "execute",
                      ("let",
                       ("var", "MemoryDescriptor"),
                       ("var", "permissions"))),
                     ("in", "write",
                      ("let",
                       ("var", "MemoryDescriptor"),
                       ("var", "permissions")))))

        # Two binary intersections become one variadic intersection and the
        # let-forms now have a Binding as their LHS whenever possible.
        expected = ("&",
                    ("let",
                     ("var", "MemoryDescriptor"),
                     ("let",
                      ("var", "process"),
                      ("==",
                       ("let", ("var", "Process"), ("var", "command")),
                       "Adium"))),
                    ("in", "execute",
                     ("let",
                      ("var", "MemoryDescriptor"),
                      ("var", "permissions"))),
                    ("in", "write",
                     ("let",
                      ("var", "MemoryDescriptor"),
                      ("var", "permissions"))))

        self.assertTransform("normalizer", original, expected)



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER abstract syntax.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"


class EfilterError(Exception):
    query = None
    root = None
    message = None
    start = None
    end = None

    def __init__(self, query, message, root=None, start=None, end=None):
        super(EfilterError, self).__init__(message)

        self.query = query
        self.message = message
        self.root = root
        self.start = start if start is not None else self.root.start
        self.end = end if end is not None else self.root.end

    def __str__(self):
        if self.start is not None and self.end is not None:
            query = "%s >>> %s <<< %s" % (
                self.query.source[0:self.start],
                self.query.source[self.start:self.end],
                self.query.source[self.end:])
        else:
            query = self.query.source

        return "%s (%s) in query %r" % (
            type(self).__name__,
            self.message,
            query)

    def __repr__(self):
        return "%s(message=%r, start=%r, end=%r)" % (
            type(self), self.message, self.start, self.end)


class EfilterParseError(EfilterError):
    token = None

    def __init__(self, *args, **kwargs):
        self.token = kwargs.pop("token")
        super(EfilterParseError, self).__init__(*args, **kwargs)


class EfilterTypeError(EfilterError):
    expected = None
    actual = None

    def __init__(self, *args, **kwargs):
        self.expected = kwargs.pop("expected")
        self.actual = kwargs.pop("actual")

        if self.expected and self.actual:
            kwargs.setdefault("message", "Expected type %r, got %r instead." %
                              (self.expected, self.actual))

        super(EfilterTypeError, self).__init__(*args, **kwargs)



import unittest

from efilter import query


class RuleAnalyzerTest(unittest.TestCase):
    def analyzeQuery(self, source):
        q = query.Query(source, syntax="dotty")
        return q.run_engine("analyzer")

    def analyzeLegacyQuery(self, source):
        q = query.Query(source, syntax="dotty")
        return q.run_engine("analyzer")

    def testBasic(self):
        analysis = self.analyzeQuery("ProcessName == 'init'")
        self.assertIn("ProcessName", analysis.symbols)

    def testMembership(self):
        analysis = self.analyzeQuery("ProcessName in ('init', 'launchd')")
        self.assertIn("ProcessName", analysis.symbols)
        self.assertIn("ProcessName", analysis.eq_indexables)

    def testScoped(self):
        analysis = self.analyzeQuery("Process.name == 'foo'")
        self.assertIn("Process", analysis.symbols)
        self.assertIn("Process.name", analysis.symbols)
        self.assertIn("Process.name", analysis.eq_indexables)

    def testNested(self):
        analysis = self.analyzeQuery(
            "Process.parent where (Process.name == 'init')")
        self.assertIn("Process.parent", analysis.symbols)
        self.assertIn("Process.name", analysis.symbols)



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Primitive types not found in Python.

This module implements a set-like container that supports multiple hashable
elems per entry.
"""


__author__ = "Adam Sindelar <adamsh@google.com>"


from efilter.protocols import indexable


class IndexSet(object):
    _backing_dict = None
    _elem_count = None

    def __init__(self, elems=()):
        self._backing_dict = dict()
        self._elem_count = 0
        for elem in elems:
            self.add(elem)

    def add(self, elem):
        duplicate_entry = False
        for index in indexable.indices(elem):
            if index in self._backing_dict:
                duplicate_entry = True

            self._backing_dict[index] = elem

        if not duplicate_entry:
            self._elem_count += 1

    def get(self, elem):
        indices = getattr(elem, "indices", None)
        if not indices:
            return None

        for index in indices:
            result = self._backing_dict.get(index)
            if result:
                return result

        return None

    def remove(self, elem):
        indices = getattr(elem, "indices", None)
        if not indices:
            raise KeyError("%s elem is not in %s.", repr(elem), repr(self))

        for index in elem.indices:
            del self._backing_dict[index]

        self._elem_count -= 1

    def discard(self, elem):
        try:
            self.remove(elem)
        except KeyError:
            return

    def pop(self):
        popped_elem = None
        for elem in self._backing_dict.itervalues():
            popped_elem = elem
            break

        self.remove(popped_elem)
        return popped_elem

    def clear(self):
        self._backing_dict = dict()
        self._elem_count = 0

    def get_index(self, index):
        return self._backing_dict.get(index)

    @property
    def indices(self):
        return self._backing_dict.keys()

    @property
    def values(self):
        return list(self)

    def isdisjoint(self, other):
        return not self.intersection(other)

    def issubset(self, other):
        for elem in self:
            if not elem in other:
                return False

        return True

    def issuperset(self, other):
        return other.issubset(self)

    def update(self, other):
        for elem in other:
            if elem in self:
                continue

            self.add(elem)

        return self

    def union(self, other):
        result = type(self)(self._backing_dict.itervalues())
        for elem in other:
            if elem in result:
                continue

            result.add(elem)

        return result

    def intersection(self, other):
        result = type(self)()
        for elem in self:
            if elem in other:
                result.add(elem)

        return result

    def intersection_update(self, other):
        for elem in list(self):
            if not elem in other:
                self.remove(elem)

        return self

    def difference(self, other):
        result = type(self)()
        for elem in self:
            if not elem in other:
                result.add(elem)

        return result

    def difference_update(self, other):
        for elem in other:
            self.remove(elem)

        return self

    def __contains__(self, elem):
        if self.get(elem) is not None:
            return True

        return False

    def __repr__(self):
        return "FrozenSet(%s)" % ", ".join([repr(elem) for elem in self])

    def __len__(self):
        return self._elem_count

    def __iter__(self):
        seen = set()

        for elem in self._backing_dict.itervalues():
            indices = set(indexable.indices(elem))
            if seen & indices:
                continue

            yield elem
            seen |= indices

    def __eq__(self, other):
        return sorted(self.indices) == sorted(other.indices)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __nonzero__(self):
        return self._elem_count > 0

    def __lt__(self, other):
        return self != other and self.issubset(other)

    def __gt__(self, other):
        return self != other and self.issuperset(other)

    def __le__(self, other):
        return self.issubset(other)

    def __ge__(self, other):
        return self.issuperset(other)

    def __or__(self, other):
        return self.union(other)

    def __sub__(self, other):
        return self.difference(other)

    def __and__(self, other):
        return self.intersection(other)

    def __ior__(self, other):
        return self.update(other)

    def __iand__(self, other):
        return self.intersection_update(other)

    def __isub__(self, other):
        return self.difference_update(other)

indexable.IIndexable.implemented(for_type=IndexSet)



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER type system.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

from efilter.protocols import indexable
from efilter.protocols import hashable
from efilter.protocols import superposition

from efilter.ext import indexset


# pylint: disable=protected-access
class DelegatingSuperposition(object):
    """Superposition implementation using a container type (set or similar.)"""

    _state_type = None
    _delegate = None

    def __init__(self, first_state=None, *states):
        self._delegate = self._make_delegate()

        if first_state is None:
            return

        self._state_type = superposition.state_type(first_state)
        self.add_state(first_state)

        for state in states:
            if superposition.state_type(state) != self.state_type():
                raise TypeError(
                    "All states of a superposition must be of the same type."
                    " First argument was of type %r, but argument %r is of "
                    " type %r." %
                    (self.state_type(), state,
                     superposition.state_type(state)))
            self.add_state(state)

    def _make_delegate(self):
        """Instantiate the container we store states in and return it.

        Currently, the container must behave like a Python set, in that it
        supports all the standard Python set opetations.
        """
        raise NotImplementedError("Subclasses must override.")

    def add_state(self, state):
        """Add state to this superposition.

        WARNING: this mutates the object (it's NOT copy on write). Unless
        you're absolutely certain of what you're doing, you most likely want
        to call superposition.state_union(sp, state) instead.
        """
        for state_ in superposition.getstates(state):
            self._add_state(state_)

    def _add_state(self, state):
        """Add state to self._delegate."""
        raise NotImplementedError("Subclasses must override.")

    def getstates(self):
        return self._delegate

    def hasstate(self, state):
        return state in self._delegate

    def state_type(self):
        return self._state_type

    def state_eq(self, other):
        if isinstance(other, type(self)):
            return self._delegate == other._delegate

        return sorted(self._delegate) == sorted(superposition.getstates(other))

    def __eq__(self, other):
        if not isinstance(other, superposition.ISuperposition):
            return False

        return self.state_eq(other)

    def __ne__(self, other):
        return not self == other

    def indices(self):
        raise NotImplementedError("Subclasses must override.")

    def _typecheck(self, other, operation="compare"):
        other_type = superposition.state_type(other)
        if other_type != self._state_type:
            raise TypeError(
                "Can't %s %r of state type %r with %r of state type %r."
                % (operation, other, other_type, self, self._state_type))

    def union(self, other):
        self._typecheck(other, "join")

        other_states = self._make_delegate()
        other_states.update(superposition.getstates(other))
        other_states.update(self._delegate)

        return type(self)(*other_states)

    def intersection(self, other):
        self._typecheck(other, "intersect")

        other_states = self._make_delegate()
        other_states.update(superposition.getstates(other))

        return type(self)(*(self._delegate & other_states))

    def difference(self, other):
        self._typecheck(other, "subtract")

        other_states = self._make_delegate()
        new_states = self._delegate - other_states

        return type(self)(*new_states)

    def issuperset(self, other):
        if isinstance(other, type(self)):
            return self._delegate >= other._delegate

        if not superposition.insuperposition(other):
            return self.hasstate(superposition.getstate(other))

        other_states = self._make_delegate()
        other_states.update(superposition.getstates(other))

        return self._delegate >= other_states

    def apply(self, f):
        return superposition.superposition(*[f(x) for x in self.getstates()])

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__,
                           ", ".join([repr(s) for s in self.getstates()]))


superposition.ISuperposition.implement(
    for_type=DelegatingSuperposition,
    implementations={
        superposition.getstates: DelegatingSuperposition.getstates,
        superposition.state_type: DelegatingSuperposition.state_type,
        superposition.state_union: DelegatingSuperposition.union,
        superposition.state_intersection: DelegatingSuperposition.intersection,
        superposition.state_difference: DelegatingSuperposition.difference,
        superposition.hasstate: DelegatingSuperposition.hasstate,
        superposition.state_eq: DelegatingSuperposition.state_eq,
        superposition.state_superset: DelegatingSuperposition.issuperset,
        superposition.state_apply: DelegatingSuperposition.apply
    }
)


indexable.IIndexable.implement(
    for_type=DelegatingSuperposition,
    implementations={
        indexable.indices: DelegatingSuperposition.indices
    }
)


class HashedSuperposition(DelegatingSuperposition):
    def _make_delegate(self):
        return set()

    def _add_state(self, state):
        self._delegate.add(state)

    def indices(self):
        result = set()
        for state in self._delegate:
            result.add(state)

        return result

superposition.superposition.implement(
    for_type=hashable.IHashable,
    implementation=HashedSuperposition)


class IndexedSuperposition(DelegatingSuperposition):
    def _make_delegate(self):
        return indexset.IndexSet()

    def _add_state(self, state):
        self._delegate.add(state)

    def indices(self):
        result = set()
        for state in self._delegate:
            result |= frozenset(indexable.indices(state))

        return result

superposition.superposition.implement(
    for_type=indexable.IIndexable,
    implementation=IndexedSuperposition)


# Make sure we're hashing anything we can hash, because it's faster.
superposition.superposition.prefer_type(
    hashable.IHashable,
    over=indexable.IIndexable)



# -*- coding: utf-8 -*-

# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""EFILTER abstract type system.

This special protocol defines functions for an application/service level
delegate object, intended to provide type information and a global name service
to EFILTER. Implementing this protocol will enable EFILTER expressions to
reference globals and provide stronger type protections and hints in the
query analyzer.
"""

from efilter import dispatch
from efilter import protocol


# Declarations:
# pylint: disable=unused-argument


@dispatch.polymorphic
def reflect(delegate, name, scope=None):
    """Provide the type of 'name', which is either a global or part of 'scope'.

    Arguments:
        delegate: The application delegate.
        name: The name to be reflected. Either a global (from getnames) or
            a member of 'scope', which is a type or a container.
        scope: The optional type scope, which was returned by a previous call
            to 'reflect'. Used to qualify that 'name' is not global, but a
            member of 'scope'.

    Returns:
        The type of 'name' or None. Invalid names should return None,
        whereas valid names with unknown type should return AnyType.

    Examples:
        # What's a process?
        reflect(delegate, "Process") #=> Process

        # Invalid - must be scoped.
        reflect(delegate, "name") #=> None

        # Attribute 'ProcessName' of a Process Entity.
        reflect(delegate, "name", Process) #=> basestring

        # Plugin output:
        reflect(delegate, "pslist") #=> pslist
        reflect(delegate, "pid", pslist) #=> int
        reflect(delegate, "process", netstat) #=> Entity

        # Globals:
        reflect(delegate, "_PsListHead") #=> proc
        reflect(delegate, "p_pid", proc) #=> unsigned
        reflect(delegate, "void_pointer", proc) #=> AnyType
    """
    raise NotImplementedError()


@dispatch.polymorphic
def asglobal(delegate, name, scope=None):
    """Does the name (with scope) represent a global?

    Arguments:
        delegate: The application delegate.
        name: The name to be reduced to global, if possible.
        scope: The scope the name was found in.

    Returns:
        None if the name is not a global. Otherwise the global name.

    Examples:
        asglobal(delegate, "Process") #=> "Process"
        asglobal(delegate, "name", Process) #=> None
        asglobal(delegate, "Process", )
    """
    raise NotImplementedError()


@dispatch.polymorphic
def provide(delegate, name):
    """Provide the value of a global constant.

    Examples:
        # Global constant:
        provide(delegate, "_PsListHead") #=> <struct pslisthead @0xffff503ac...
    """
    raise NotImplementedError()


@dispatch.polymorphic
def getnames(delegate):
    """Provide a list of global names which can be reflected or provided.

    Returns an iterable of strings.
    """
    raise NotImplementedError()


class INameDelegate(protocol.Protocol):
    _protocol_functions = (reflect, provide, getnames)



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER engine base class.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"


import abc

from efilter.protocols import name_delegate


class NopAppDelegate(object):
    """This application delegate doesn't do anything."""


name_delegate.INameDelegate.implement(
    for_type=NopAppDelegate,
    implementations={
        name_delegate.reflect: lambda d, n, s: None,
        name_delegate.provide: lambda d, n: None,
        name_delegate.getnames: lambda d: ()
    }
)


class Engine(object):
    """Base class representing the various behaviors of the EFILTER AST.

    Attributes:
        query: The query object this is running on.
        application_delegate: The application delegate - should be the same as
            the one on the query.
    """
    __metaclass__ = abc.ABCMeta

    ENGINES = {}

    def __init__(self, query, application_delegate=None):
        super(Engine, self).__init__()
        self.query = query
        if application_delegate:
            self.application_delegate = application_delegate
        else:
            self.application_delegate = NopAppDelegate()

    @abc.abstractmethod
    def run(self, *_, **__):
        """Execute this engine and return its result."""

    @classmethod
    def register_engine(cls, subcls, shorthand=None):
        if shorthand is None:
            shorthand = repr(subcls)

        cls.ENGINES[shorthand] = subcls

    @classmethod
    def get_engine(cls, shorthand):
        if isinstance(shorthand, type) and issubclass(shorthand, Engine):
            return shorthand

        return cls.ENGINES.get(shorthand)


class VisitorEngine(Engine):
    """Engine that implements the visitor pattern.

    Visitor engines start by calling self.visit(self.root). The default
    implementation of self.visit will walk the MRO of the expression it got
    and find the best available handler in the form of visit_<classname>.

    (This ends up trying visit_Expression as last resort and then throwing a
    TypeError if no handlers are available.)

    The actual handlers themselves are implemented by subclasses providing
    different behaviors (e.g. matcher, analyzer, etc.). Each handler is
    responsible for evaluating the branch it's passed, usually by recursively
    calling self.visit on any children and then combining the results of those
    sub-branches according to its own logic.

    For example, to implement expression.Sum (add up numbers), one would do:

        def visit_Sum(self, expr):
            result = 0
            for branch in expr.children:
                result += self.visit(branch)

            return result
    """

    def __hash__(self):
        return hash((type(self), self.query))

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.query == other.query

    def __ne__(self, other):
        return not self.__eq__(other)

    def run(self, *_, **kwargs):
        """Visitors by default visit the root of the query."""
        super(VisitorEngine, self).run()
        self.node = self.query.root
        return self.visit(self.node, **kwargs)

    def fall_through(self, node, engine_shorthand, **kwargs):
        """A visitor only implementing part of the AST can delegate with this.

        If a visitor engine only implemenets a subset of the AST language, it
        becomes useful to delegate to other visitor classes for subbranches the
        original visitor cannot handle.

        Arguments:
            node: The subbranch to be handled by the other visitor.
            engine_shorthand: The shorthand name of the other visitor.
            kwargs: Are passed on verbatim.

        Returns:
            Whatever the other visitor engine returns.
        """
        engine_cls = type(self).get_engine(engine_shorthand)
        if not issubclass(engine_cls, VisitorEngine):
            raise TypeError(
                "VisitorEngine can only fall through to other VisitorEngines.")

        engine = engine_cls(query=self.query,
                            application_delegate=self.application_delegate)
        return engine.visit(node, **kwargs)

    def visit(self, node, **kwargs):
        """Visit the AST node by calling as specific a handler as we have."""
        # Walk the MRO and try to find a closest match for handler.
        for cls in type(node).mro():
            handler_name = "visit_%s" % cls.__name__
            handler = getattr(self, handler_name, None)

            if callable(handler):
                return handler(node, **kwargs)

        # No appropriate handler for this class. Explode.
        raise ValueError(
            "Visitor engine %s has no handler for node %r of %r." %
            (type(self).__name__, node, self.query))



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Lisp-like EFILTER syntax.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

from efilter import expression
from efilter import frontend


EXPRESSIONS = {
    "var": expression.Binding,
    "!": expression.Complement,
    "let": expression.Let,
    "let-any": expression.LetAny,
    "let-each": expression.LetEach,
    "in": expression.Membership,
    "regex": expression.RegexFilter,
    "|": expression.Union,
    "&": expression.Intersection,
    ">": expression.StrictOrderedSet,
    ">=": expression.PartialOrderedSet,
    "==": expression.Equivalence,
    "+": expression.Sum,
    "-": expression.Difference,
    "*": expression.Product,
    "/": expression.Quotient,
}


class Parser(frontend.Frontend):
    """Parses the lisp expression language into the query AST."""

    @property
    def root(self):
        return self._parse_atom(self.original)

    def _parse_atom(self, atom):
        if isinstance(atom, tuple):
            return self._parse_s_expression(atom)

        return expression.Literal(atom)

    def _parse_s_expression(self, atom):
        car = atom[0]
        cdr = atom[1:]

        # Bindings are a little special.
        if car == "var":
            return expression.Binding(cdr[0])

        return EXPRESSIONS[car](*(self._parse_atom(a) for a in cdr))


frontend.Frontend.register_frontend(Parser, shorthand="lisp")



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER test helpers.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

import unittest

from efilter import expression
from efilter import protocol
from efilter import query


class EngineTestCase(unittest.TestCase):
    def _guess_syntax(self, source, default=None):
        if default:
            return default

        if isinstance(source, expression.Expression):
            return "expression"

        if isinstance(source, basestring):
            return "dotty"

        if isinstance(source, tuple):
            return "lisp"

        raise ValueError("Cannot guess syntax of %r." % source)

    def parse_query(self, source, source_syntax=None, app_delegate=None):
        source_syntax = self._guess_syntax(source, source_syntax)
        return query.Query(source, syntax=source_syntax,
                           application_delegate=app_delegate)

    def get_engine_result(self, engine, source, source_syntax=None,
                          app_delegate=None, **kwargs):
        q = self.parse_query(source=source, source_syntax=source_syntax,
                             app_delegate=app_delegate)
        return q.run_engine(engine, **kwargs)

    # Custom assertions:

    def assertEngineResult(self, engine, source, expected, source_syntax=None,
                           assertion=None, app_delegate=None, **kwargs):
        result = self.get_engine_result(
            engine, source, source_syntax, app_delegate, **kwargs)
        if callable(assertion):
            assertion(expected, result)
        else:
            self.assertEqual(expected, result)

    def assertTransform(self, engine, original, expected, original_syntax=None,
                        expected_syntax=None, app_delegate=None, **kwargs):
        orig_query = self.parse_query(original, original_syntax, app_delegate)
        actual = orig_query.run_engine(engine, **kwargs)
        baseline = self.parse_query(expected, expected_syntax, app_delegate)

        self.assertEqual(actual, baseline)

    def assertIsa(self, t, p):
        self.assertTrue(protocol.isa(t, p))



import unittest

from efilter.protocols import hashable
from efilter.protocols import superposition


# Do not try any of this in real code - this is just for sake of tests.
class frozendict(dict):
    def __hash__(self):
        return hash(frozenset(self.items()))


def _hashed_frozendict(x):
    return hash(frozenset(x.items()))

hashable.IHashable.implement(
    for_type=frozendict,
    implementations={hashable.hashed: _hashed_frozendict})


class SuperpositionTest(unittest.TestCase):
    def assertStateEq(self, s1, s2):
        return self.assertTrue(superposition.state_eq(s1, s2))

    def testMutables(self):
        """Test that mutable types cannot be put in a superposition."""
        x = dict(name="Alice")
        y = dict(name="Bob")

        # Can't put dicts in a superposition.
        with self.assertRaises(NotImplementedError):
            superposition.superposition(x, y)

        # Frozendicts work, because they are hashable.
        x = frozendict(name="Alice")
        y = frozendict(name="Bob")
        superposition.superposition(x, y)

    def testCreation(self):
        """Test that creation is reasonable."""
        
        # Providing the same object twice will still build a superposition.
        s = superposition.superposition("foo", "foo")
        # This object is a superposition type...
        self.assertIsInstance(s, superposition.ISuperposition)

        # ...but it is not IN superposition.
        self.assertFalse(superposition.insuperposition(s))

        # Using meld is sometimes more convenient for this.
        s = superposition.meld("foo", "foo")
        # This object is actually a string.
        self.assertIsInstance(s, basestring)
        # It can still be manipulated with the superposition-aware protocol,
        # as can any scalar.
        self.assertEqual(s, superposition.getstate(s))

    def testTypes(self):
        """Test that types are correctly derived and enforced."""

        # Cannot have a superposition with two different state types.
        with self.assertRaises(TypeError):
            superposition.superposition(1, "foo")

        self.assertEqual(superposition.superposition(1, 2).state_type(), int)

    def testStates(self):
        """Test that states are inspectable and comparable."""
        s1 = superposition.superposition("foo", "bar")
        s2 = superposition.superposition("bar", "foo")
        s3 = superposition.superposition(1, 2)
        s4 = 1
        s5 = superposition.superposition(1)

        self.assertItemsEqual(superposition.getstates(s1),
                              superposition.getstates(s2))

        self.assertTrue(superposition.state_eq(s1, s2))
        self.assertFalse(superposition.state_eq(s1, s3))

        # Superposition is obviously not equal to a scalar.
        self.assertFalse(s5 == s4)

        # But their states CAN be equal:
        self.assertTrue(superposition.state_eq(s4, s5))
        self.assertTrue(superposition.state_eq(s5, s4))

        # We can also compare two scalars this way (if we really have nothing
        # better to do).
        self.assertTrue(superposition.state_eq("foo", "foo"))

    def testSetOps(self):
        """Test set operations on states."""
        s1 = 1
        s2 = superposition.superposition(1, 2, 3)
        s3 = superposition.superposition(2, 3, 4)
        s4 = 2

        # Test intersections:
        s_ = superposition.state_intersection(s1, s2)
        self.assertStateEq(s_, 1)

        s_ = superposition.state_intersection(s1, s3)
        self.assertStateEq(s_, None)

        s_ = superposition.state_intersection(s2, s3)
        self.assertStateEq(superposition.superposition(2, 3), s_)

        # Test unions:
        s_ = superposition.state_union(s1, s4)
        self.assertStateEq(superposition.superposition(1, 2), s_)

    def testNesting(self):
        """Test that superpositions remain flat."""
        s = superposition.superposition("foo", "bar")
        s = superposition.superposition(s, "baz")
        self.assertStateEq(superposition.superposition("foo", "bar", "baz"), s)

        s = superposition.superposition("zoo", s)
        self.assertStateEq(
            superposition.superposition("foo", "bar", "baz", "zoo"), s)

        self.assertEqual(superposition.state_type(s), type("foo"))

    def testApplication(self):
        """Test function application across states."""
        self.assertEqual(
            superposition.superposition(2, 4),
            superposition.state_apply(
                superposition.superposition(1, 2),
                lambda x: x * 2))

        # As everything working on states, this should also work on scalars.
        applied = superposition.state_apply(5, lambda x: x * 2)
        self.assertTrue(superposition.state_eq(10, applied))

    def testHasState(self):
        """Test superposition operations on scalars."""
        # hasstate should work same as equivalance on scalars.
        self.assertTrue(superposition.hasstate(1, 1))
        self.assertFalse(superposition.hasstate(2, 1))

        with self.assertRaises(TypeError):
            # The 'state' argument to hasstate is not allowed to be in
            # superposition.
            superposition.hasstate(1, superposition.superposition(1, 2))

        # It should work as expected in the base case:
        self.assertTrue(superposition.hasstate(
            superposition.superposition("foo", "bar"),
            state="foo"))
        self.assertFalse(superposition.hasstate(
            superposition.superposition("foo", "bar"),
            state="baz"))



import unittest

from efilter import query


class DottyOutputTest(unittest.TestCase):
    def assertOutput(self, original, output):
        q = query.Query(original, syntax="dotty")
        actual_output = q.run_engine("dotty_output")
        self.assertEqual(output, actual_output)

    def testBasic(self):
        self.assertOutput(original="5 + 5 == 10",
                          output="5 + 5 == 10")

    def testSimpleLet(self):
        self.assertOutput(original="Process.name == 'foo'",
                          output="Process.name == 'foo'")

    def testWhere(self):
        self.assertOutput(
            original="Process.parent where (name == 'foo' and pid == 5)",
            output="Process.parent where (name == 'foo' and pid == 5)")

    def testAnywhere(self):
        self.assertOutput(
            original="any Process.parent where (name == 'foo')",
            output="any Process.parent where (name == 'foo')")



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER dotty syntax output.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

from efilter import engine
from efilter import expression

from efilter.frontends.experiments import dotty


def build_operator_lookup(*tables):
    lookup = {}
    for table in tables:
        for token, operator in table.iteritems():
            if not (isinstance(operator.handler, type) and
                    issubclass(operator.handler, expression.Expression)):
                continue

            lookup[operator.handler] = token

    return lookup


class DottyOutput(engine.VisitorEngine):
    """Produces equivalent Dotty output to the AST.

    This class follows the visitor pattern. See documentation on VisitorEngine.
    """

    TOKENS = build_operator_lookup(dotty.INFIX, dotty.PREFIX)

    def visit_Let(self, expr):
        lhs = expr.lhs
        rhs = expr.rhs
        left = self.visit(lhs)
        right = self.visit(rhs)
        token = "."

        if not isinstance(expr.lhs, (expression.ValueExpression,
                                     expression.Let)):
            left = "(%s)" % left
            token = " where "

        if not isinstance(expr.rhs, (expression.ValueExpression,
                                     expression.Let)):
            right = "(%s)" % right
            token = " where "

        return token.join((left, right))

    def visit_LetAny(self, expr):
        return "any %s" % self.visit_Let(expr)

    def visit_letEach(self, expr):
        return "each %s" % self.visit_Let(expr)

    def visit_Literal(self, expr):
        return repr(expr.value)

    def visit_Binding(self, expr):
        return expr.value

    def visit_Complement(self, expr):
        return "not (%s)" % self.visit(expr)

    def visit_BinaryExpression(self, expr):
        return self.visit_VariadicExpression(expr)

    def visit_VariadicExpression(self, expr):
        token = self.TOKENS[type(expr)]
        separator = " %s " % token
        return separator.join(self.visit(x) for x in expr.children)


engine.Engine.register_engine(DottyOutput, "dotty_output")



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER versioning.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

import datetime
from dateutil import parser
import pytz
import re
import subprocess


def _unix_epoch(date):
    td =  date - datetime.datetime(1970, 1, 1, tzinfo=pytz.UTC)
    return int(td.total_seconds())


def run_git_log():
    """Generate version based on date of last commit."""
    try:
        p = subprocess.Popen(
            ["git", "log", "-1", "--format=%cd", "--date=iso-strict"],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False)
        p.stderr.close()
        output = p.stdout.readlines()[0]
        date = parser.parse(output)
        return _unix_epoch(date)
    except OSError:
        return None


def get_pkg_version():
    """Get version string by parsing PKG-INFO."""
    try:
        with open("PKG-INFO", "r") as fp:
            rgx = re.compile(r"Version: (\d+\.\d+\.\d+)")
            for line in fp.readlines():
                match = rgx.match(line)
                if match:
                    return match.group(1)
    except IOError:
        return None


def get_version():
    """Get a version string as date of last commit or else parse PKG-INFO.

    Example:
        2015.07.10
    """
    version = run_git_log()
    if version:
        return version

    version = get_pkg_version()
    if version:
        return version

    raise RuntimeError("Couldn't get git log or PKG-INFO to guess version.")


# -*- coding: utf-8 -*-

# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""EFILTER abstract type system."""

from efilter import dispatch
from efilter import protocol


# Declarations:
# pylint: disable=unused-argument


@dispatch.polymorphic
def sum(x, y):
    raise NotImplementedError()


@dispatch.polymorphic
def product(x, y):
    raise NotImplementedError()


@dispatch.polymorphic
def difference(x, y):
    raise NotImplementedError()


@dispatch.polymorphic
def quotient(x, y):
    raise NotImplementedError()


class INumber(protocol.Protocol):
    _protocol_functions = (sum, product, difference, quotient)


# Default implementations:

INumber.implement(
    for_types=(int, float, complex, long),
    implementations={
        sum: lambda x, y: x + y,
        product: lambda x, y: x * y,
        difference: lambda x, y: x - y,
        quotient: lambda x, y: x / y
    }
)



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Canonical EFILTER syntax parser.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"


import collections
import re

from efilter import errors
from efilter import expression
from efilter import frontend


# Transformation functions, for expressions that don't directly map to
# something in the AST.

def ComplementEquivalence(*args, **kwargs):
    return expression.Complement(
        expression.Equivalence(*args, **kwargs), **kwargs)


def ComplementMembership(*args, **kwargs):
    return expression.Complement(
        expression.Membership(*args, **kwargs), **kwargs)


def ReverseStrictOrderedSet(*args, **kwargs):
    return expression.StrictOrderedSet(*reversed(args), **kwargs)


def ReversePartialOrderedSet(*args, **kwargs):
    return expression.PartialOrderedSet(*reversed(args), **kwargs)


def NegateValue(*args, **kwargs):
    return expression.Product(
        expression.Literal(-1),
        *args,
        **kwargs)


def FlattenComponentLiteral(*args, **kwargs):
    if not isinstance(args[0], expression.Binding):
        raise ValueError(
            "'has component' must be followed by a component. Got %s." % (
                args[0]))
    return expression.ComponentLiteral(args[0].value, **kwargs)


def TransformLetAny(let, **kwargs):
    if not isinstance(let, expression.Let):
        raise ValueError("'any' must be followed by a 'matches' expression.")
    context, expr = let.children
    return expression.LetAny(context, expr, **kwargs)


def TransformLetEach(let, **kwargs):
    if not isinstance(let, expression.Let):
        raise ValueError("'each' must be followed by a 'matches' expression.")
    context, expr = let.children
    return expression.LetEach(context, expr, **kwargs)


# Operators - infix and prefix.

Operator = collections.namedtuple("Operator",
                                  "precedence assoc handler docstring")

# The order of precedence matters for generated matching rules, which is why
# this is an OrderedDict.
INFIX = collections.OrderedDict([
    ("->", Operator(precedence=2, assoc="left", handler=expression.Let,
                    docstring="Shorthand for 'matches'.")),
    ("+", Operator(precedence=3, assoc="left", handler=expression.Sum,
                   docstring="Arithmetic addition.")),
    ("-", Operator(precedence=3, assoc="left", handler=expression.Difference,
                   docstring="Arithmetic subtraction.")),
    ("*", Operator(precedence=5, assoc="left", handler=expression.Product,
                   docstring="Arithmetic multiplication.")),
    ("/", Operator(precedence=5, assoc="left", handler=expression.Quotient,
                   docstring="Arithmetic division.")),
    ("==", Operator(precedence=2, assoc="left",
                    handler=expression.Equivalence,
                    docstring="Equivalence (same as 'is').")),
    ("!=", Operator(precedence=2, assoc="left",
                    handler=ComplementEquivalence,
                    docstring="Inequivalence (same as 'is not').")),
    ("is not", Operator(precedence=2, assoc="left",
                        handler=ComplementEquivalence,
                        docstring="Inequivalence (same as '!=').")),
    ("is", Operator(precedence=2, assoc="left",
                    handler=expression.Equivalence,
                    docstring="Equivalence (same as '==')")),
    ("not in", Operator(precedence=2, assoc="left",
                        handler=ComplementMembership,
                        docstring="Left-hand operand is not in list.")),
    ("in", Operator(precedence=2, assoc="left",
                    handler=expression.Membership,
                    docstring="Left-hand operand is in list.")),
    (">", Operator(precedence=2, assoc="left",
                   handler=expression.StrictOrderedSet,
                   docstring="Greater-than.")),
    (">=", Operator(precedence=2, assoc="left",
                    handler=expression.PartialOrderedSet,
                    docstring="Equal-or-greater-than.")),
    ("<", Operator(precedence=2, assoc="left",
                   handler=ReverseStrictOrderedSet,
                   docstring="Less-than.")),
    ("<=", Operator(precedence=2, assoc="left",
                    handler=ReversePartialOrderedSet,
                    docstring="Equal-or-less-than.")),
    ("matches", Operator(precedence=2, assoc="left",
                         handler=expression.Let,
                         docstring="Left-hand operand matched subquery.")),
    ("and", Operator(precedence=1, assoc="left",
                     handler=expression.Intersection,
                     docstring="Logical AND.")),
    ("or", Operator(precedence=0, assoc="left", handler=expression.Union,
                    docstring="Logical OR.")),
    ("=~", Operator(precedence=2, assoc="left",
                    handler=expression.RegexFilter,
                    docstring="Left-hand operand matches regex.")),
])


PREFIX = {
    "not": Operator(precedence=6, assoc=None, handler=expression.Complement,
                    docstring="Logical NOT."),
    "-": Operator(precedence=4, assoc=None, handler=NegateValue,
                  docstring="Unary -."),
    "has component": Operator(precedence=7, assoc=None,
                              handler=FlattenComponentLiteral,
                              docstring="Matching entity must have component."),
    "any": Operator(precedence=1, assoc=None, handler=TransformLetAny,
                    docstring=("Following 'matches' should succeed if "
                               "any left-hand value matches.")),
    "each": Operator(precedence=1, assoc=None, handler=TransformLetEach,
                     docstring=("Following 'matches' should only "
                                "succeed if all left-hand values match.")),
}


def EnumAsPattern(strings):
    """Return a regex that'll match any of the strings in order."""
    return "(%s)" % "|".join([re.escape(x) for x in strings])


class Token(object):
    """Represents a result from the tokenizer."""

    def __init__(self, name, value, start, end):
        self.name = name
        self.value = value
        self.start = start
        self.end = end

    def __repr__(self):
        return "Token(name='%s', value='%s', start=%d, end=%d)" % (
            self.name, self.value, self.start, self.end)


class Pattern(object):
    """A token pattern.

    Args:
      state_regex: If this regular expression matches the current state this
                   rule is considered.
      regex: A regular expression to try and match from the current point.
      actions: A command separated list of method names in the Lexer to call.
      next_state: The next state we transition to if this Pattern matches.
      flags: flags to re.compile.
    """

    def __init__(self, label, state_regex, regex, actions, next_state,
                 flags=re.I):
        self.state_regex = re.compile(
            state_regex, re.DOTALL | re.M | re.S | re.U | flags)
        self.regex = re.compile(regex, re.DOTALL | re.M | re.S | re.U | flags)
        self.label = label
        self.re_str = regex

        if actions:
            self.actions = actions.split(",")
        else:
            self.actions = []

        self.next_state = next_state


class Tokenizer(object):
    """Context-free tokenizer for the efilter language.

    This is a very basic pattern-based tokenizer. Any rule from patterns
    will try to match the next token in the buffer if its state_regex matches
    the current state. Only meaningful tokens are emitted (not whitespace.)
    """
    patterns = [
        # Keywords, operators and symbols
        Pattern("symbol", "INITIAL", r"(\&?[A-Z][A-Za-z0-9]+\/[a-z0-9_]+)",
                "emit", None),
        Pattern("infix", "INITIAL", EnumAsPattern(INFIX.keys()),
                "emit", None),
        Pattern("prefix", "INITIAL", EnumAsPattern(PREFIX.keys()),
                "emit", None),
        Pattern("lparen", "INITIAL", r"\(",
                "emit", None),
        Pattern("rparen", "INITIAL", r"\)",
                "emit", None),
        Pattern("comma", "INITIAL", r",",
                "emit", None),
        Pattern("symbol", "INITIAL", r"([a-z_\.][a-z_\.0-9]+)", "emit", None),
        Pattern("param", "INITIAL", r"\{([a-z_0-9]*)\}", "emit_param", None),

        # Numeric literals
        Pattern("literal", "INITIAL", r"(\d+\.\d+)", "emit_float", None),
        Pattern("literal", "INITIAL", r"(0x[0-9a-zA-Z]+)", "emit_int16", None),
        Pattern("literal", "INITIAL", r"(\d+)", "emit_int", None),

        # String literals
        Pattern(None, "INITIAL", r"\"", "string_start", "STRING"),
        Pattern(None, "INITIAL", r"'", "string_start", "SQ_STRING"),

        Pattern("literal", "STRING", "\"", "pop_state,emit_string", None),
        Pattern(None, "STRING", r"\\(.)", "string_escape", None),
        Pattern(None, "STRING", r"[^\\\"]+", "string_append", None),

        Pattern("literal", "SQ_STRING", "'", "pop_state,emit_string", None),
        Pattern(None, "SQ_STRING", r"\\(.)", "string_escape", None),
        Pattern(None, "SQ_STRING", r"[^\\']+", "string_append", None),

        # Whitespace is ignored.
        Pattern(None, ".", r"\s+", None, None),
    ]

    def __init__(self, query):
        self.buffer = query
        self.state_stack = ["INITIAL"]
        self.current_token = None
        self._position = 0
        self.limit = len(query)
        self.lookahead = []
        self._param_idx = 0

    @property
    def position(self):
        """Returns the logical position (unaffected by lookahead)."""
        if self.lookahead:
            return self.lookahead[0].start

        return self._position

    def pop_state(self, **_):
        try:
            self.state_stack.pop()
        except IndexError:
            self.error("Pop state called on an empty stack.", self.position)

    def next_token(self):
        """Returns the next logical token.

        Will trigger parsing if it has to.
        """
        if self.lookahead:
            self.current_token = self.lookahead.pop(0)
            return self.current_token

        self.current_token = self._parse_next_token()
        return self.current_token

    def _parse_next_token(self):
        """Will parse patterns until it gets to the next token or EOF."""
        while self._position < self.limit:
            token = self.next_pattern()
            if token:
                return token

        return None

    def peek(self, steps=1):
        """Look ahead, doesn't affect current_token and next_token."""
        while len(self.lookahead) < steps:
            token = self._parse_next_token()
            if token is None:
                return None

            self.lookahead.append(token)

        return self.lookahead[steps - 1]

    def parse(self):
        """Yield every token in turn."""
        while self._position < self.limit:
            token = self.next_token()
            if not token:
                return

            yield token

    def next_pattern(self):
        """Parses the next pattern by matching each in turn."""
        current_state = self.state_stack[-1]
        position = self._position
        for pattern in self.patterns:
            if not pattern.state_regex.match(current_state):
                continue

            m = pattern.regex.match(self.buffer, position)
            if not m:
                continue

            position = m.end()
            token = None

            if pattern.next_state:
                self.state_stack.append(pattern.next_state)

            for action in pattern.actions:
                callback = getattr(self, action, None)
                if callback is None:
                    raise RuntimeError(
                        "No method defined for pattern action %s!" % action)

                token = callback(string=m.group(0), match=m, pattern=pattern)

            self._position = position

            return token

        self.error("Don't know how to match next. Did you forget quotes?",
                   self.position)

    def error(self, message, start, end=None):
        """Print a nice error."""
        raise errors.EfilterParseError(
            query=self.buffer, start=start, end=end, message=message)

    def emit(self, string, match, pattern, **_):
        """Emits a token using the current pattern match and pattern label."""
        return Token(name=pattern.label, value=string, start=match.start(),
                     end=match.end())

    def emit_param(self, match, pattern, **_):
        param_name = match.group(1)
        if not param_name:
            param_name = self._param_idx
            self._param_idx += 1

        return Token(name=pattern.label, value=param_name, start=match.start(),
                     end=match.end())

    def emit_int(self, string, match, pattern, **_):
        return Token(name=pattern.label, value=int(string), start=match.start(),
                     end=match.end())

    def emit_int16(self, string, match, pattern, **_):
        return Token(name=pattern.label, value=int(string, 16),
                     start=match.start(), end=match.end())

    def emit_float(self, string, match, pattern, **_):
        return Token(name=pattern.label, value=float(string),
                     start=match.start(), end=match.end())

    # String parsing

    def string_start(self, match, **_):
        self.string = ""
        self.string_position = match.start()

    def string_escape(self, string, match, **_):
        if match.group(1) in "'\"rnbt":
            self.string += string.decode("string_escape")
        else:
            self.string += string

    def string_append(self, string="", **_):
        self.string += string

    def emit_string(self, pattern, match, **_):
        return Token(name=pattern.label, value=self.string,
                     start=self.string_position, end=match.end())


class Parser(frontend.Frontend):
    """Parses the efilter language into the query AST.

    This is a basic precedence-climbing parser with support for prefix
    operators and a few special cases for list literals and such.
    """

    def __init__(self, original, params=None):
        super(Parser, self).__init__(original)

        self.tokenizer = Tokenizer(self.original)

        if isinstance(params, list):
            self.params = {}
            for idx, val in enumerate(params):
                self.params[idx] = val
        else:
            self.params = params

    def _handle_expr(self, operator, *args, **kwargs):
        try:
            return operator.handler(*args, **kwargs)
        except ValueError as e:
            return self.error(e.message,
                              start_token=args[0])

    def _replace_param(self, token):
        param_name = token.value
        value = self.params.get(param_name, None)
        if value is None:
            return self.error("No value provided for param %s" % param_name,
                              token)

        return value

    def next_atom(self):
        token = self.tokenizer.next_token()

        if token is None:
            return self.error("Unexpected end of input.")

        if token.name == "infix":
            if token.value == "-":
                # As it turns out, minus signs can be prefix operators! Who
                # knew? Certainly not the tokenizer.
                token.name = "prefix"
            else:
                return self.error("Unexpected infix operator.", token)

        if token.name == "prefix":
            operator = PREFIX[token.value]
            lhs = self.next_atom()
            rhs = self.next_expression(lhs, operator.precedence)
            return self._handle_expr(operator, rhs, start=token.start,
                                     end=rhs.end)

        if token.name == "literal":
            return expression.Literal(token.value, start=token.start,
                                      end=token.end)

        if token.name == "param":
            return expression.Literal(self._replace_param(token),
                                      start=token.start, end=token.end)

        if token.name == "symbol":
            return expression.Binding(token.value, start=token.start,
                                      end=token.end)

        if token.name == "lparen":
            # Parentheses can denote subexpressions or lists. Lists have at
            # least one comma before rparen (just like Python).
            lhs = self.next_atom()
            expr = self.next_expression(lhs, 0)
            if self.tokenizer.current_token is None:
                return self.error("End of input before closing parenthesis.",
                                  token)

            if self.tokenizer.peek().name == "comma":
                # It's a list, not an expression. Build it out as a literal.
                if not isinstance(lhs, expression.Literal):
                    return self.error(
                        "Non-literal value in list.", lhs)

                self.tokenizer.next_token()
                vals = [lhs.value]

                while (self.tokenizer.current_token and
                       self.tokenizer.current_token.name == "comma"):
                    atom = self.next_atom()
                    if not isinstance(atom, expression.Literal):
                        return self.error(
                            "Non-literal value in list", atom)
                    vals.append(atom.value)
                    self.tokenizer.next_token()

                if (self.tokenizer.current_token is None or
                        self.tokenizer.current_token.name != "rparen"):
                    self.error("Lists must end with a closing paren.",
                               self.tokenizer.current_token)

                return expression.Literal(tuple(vals), start=token.start,
                                          end=self.tokenizer.position)

            elif self.tokenizer.peek().name != "rparen":
                # We got here because there's still some stuff left to parse
                # and the next token is not an rparen. That can mean that an
                # infix operator is missing or that the parens are unmatched.
                # Decide which is more likely and raise the appropriate error.
                lparens = 1
                rparens = 0
                lookahead = 2
                while self.tokenizer.peek(lookahead):
                    if self.tokenizer.peek(lookahead).name == "lparen":
                        lparens += 1
                    elif self.tokenizer.peek(lookahead).name == "rparen":
                        rparens += 1

                    lookahead += 1

                if lparens > rparens:
                    return self.error("Ummatched left parenthesis.", token)
                else:
                    next_token = self.tokenizer.peek()
                    return self.error(
                        "Was not expecting %s here." % next_token.value,
                        next_token)

            self.tokenizer.next_token()
            return expr

        return self.error("Cannot handle token %s." % token, token)

    def next_expression(self, lhs, min_precedence):
        # This loop will spin as long as:
        # 1: There is a next token.
        # 2: It is an infix operator.
        # 3: Its precedence is higher than min_precedence.
        while self.tokenizer.peek():
            token = self.tokenizer.peek()

            if token.name != "infix":
                break

            operator = INFIX[token.value]
            if operator.precedence < min_precedence:
                break

            # We're a match - consume the next token.
            self.tokenizer.next_token()

            rhs = self.next_atom()
            next_min_precedence = operator.precedence
            if operator.assoc == "LEFT":
                next_min_precedence += 1

            # Let's see if the next infix operator (if any) is of higher
            # precedence than we are.
            while (self.tokenizer.peek() and
                   self.tokenizer.peek().name == "infix"):
                next_token = self.tokenizer.peek()
                next_operator = INFIX[next_token.value]
                if next_operator.precedence < next_min_precedence:
                    break
                rhs = self.next_expression(rhs, next_operator.precedence)

            lhs = self._handle_expr(operator, lhs, rhs, start=lhs.start,
                                    end=rhs.end)

        return lhs

    def parse(self):
        result = self.next_expression(self.next_atom(), 0)
        # If we didn't consume the whole query then raise.
        if self.tokenizer.peek():
            token = self.tokenizer.peek()
            return self.error(
                "Unexpected %s '%s'. Were you looking for an operator?" %
                (token.name, token.value),
                token)

        return result

    @property
    def root(self):
        return self.parse()

    def error(self, message, start_token=None, end_token=None):
        start = self.tokenizer.position
        end = start + 20
        if start_token:
            start = start_token.start
            end = start_token.end

        if end_token:
            end = end_token.end

        raise errors.EfilterParseError(
            query=self.original, start=start, end=end, message=message,
            token=start_token)


frontend.Frontend.register_frontend(Parser, shorthand="slashy")



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER individual object filter and matcher.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"


import re

from efilter import expression
from efilter import engine

from efilter.protocols import associative
from efilter.protocols import superposition


class ObjectMatcher(engine.VisitorEngine):
    """Given a query and bindings will evaluate the query.

    This class follows the visitor pattern. See documentation on VisitorEngine.
    """

    # run() sets this to the sort value for the latest object matched.
    latest_sort_order = None

    def run(self, bindings, match_backtrace=False):
        self.match_backtrace = match_backtrace
        self.bindings = bindings

        # The match backtrace works by keeping a list of all the branches that
        # matched and then backtracking from the latest one to be evaluated
        # to the first parent that's a relation.
        if self.match_backtrace:
            self._matched_expressions = []

        self.latest_sort_order = []
        self.node = self.query.root
        self.result = self.visit(self.node)
        self.latest_sort_order = tuple(self.latest_sort_order)

        if self.match_backtrace:
            self.matched_expression = None
            for expr in self._matched_expressions:
                if isinstance(expr, expression.Relation):
                    self.matched_expression = expr
                    break

        if self.result:
            return self

        return False

    def visit(self, expr, **_):
        result = super(ObjectMatcher, self).visit(expr)

        if self.match_backtrace and result:
            self._matched_expressions.append(expr)

        return result

    def visit_Literal(self, expr, **_):
        return expr.value

    def visit_Binding(self, expr, **_):
        return associative.select(self.bindings, expr.value)

    def visit_Let(self, expr, **_):
        saved_bindings = self.bindings
        if isinstance(expr, expression.LetAny):
            union_semantics = True
        elif isinstance(expr, expression.LetEach):
            union_semantics = False
        else:
            union_semantics = None

        if not isinstance(expr.context, expression.Binding):
            raise ValueError(
                "Left operand of Let must be a Binding expression.")

        # Context to rebind to. This is the key that will be selected from
        # current bindings and become the new bindings for ever subexpression.
        context = expr.context.value

        try:
            rebind = associative.resolve(saved_bindings, context)

            if not rebind:  # No value from context.
                return None

            if union_semantics is None:
                # This is a simple let, which does not permit superposition
                # semantics.
                if superposition.insuperposition(rebind):
                    raise TypeError(
                        "A Let expression doesn't permit superposition "
                        "semantics. Use LetEach or LetAny instead.")

                self.bindings = rebind
                return self.visit(expr.expression)

            # If we're using union or intersection semantics, the type of
            # rebind MUST be a Superposition, even if it happens to have
            # only one state. If the below throws a type error then the
            # query is invalid and should fail here.
            result = False
            for state in superposition.getstates(rebind):
                self.bindings = state
                result = self.visit(expr.expression)
                if result and union_semantics:
                    return result

                if not result and not union_semantics:
                    return False

            return result
        finally:
            self.bindings = saved_bindings

    def visit_ComponentLiteral(self, expr, **_):
        return getattr(self.bindings.components, expr.value)

    def visit_Complement(self, expr, **_):
        return not self.visit(expr.value)

    def visit_Intersection(self, expr, **_):
        for child in expr.children:
            if not self.visit(child):
                return False

        return True

    def visit_Union(self, expr, **_):
        for child in expr.children:
            if self.visit(child):
                return True

        return False

    def visit_Sum(self, expr, **_):
        return sum([self.visit(child) for child in expr.children])

    def visit_Difference(self, expr, **_):
        difference = self.visit(expr.children[0])
        for child in expr.children[1:]:
            difference -= self.visit(child)

        return difference

    def visit_Product(self, expr, **_):
        product = 1
        for child in expr.children:
            product *= self.visit(child)

        return product

    def visit_Quotient(self, expr, **_):
        quotient = self.visit(expr.children[0])
        for child in expr.children[1:]:
            quotient /= self.visit(child)

        return quotient

    def visit_Equivalence(self, expr, **_):
        first_val = self.visit(expr.children[0])
        for child in expr.children[1:]:
            if self.visit(child) != first_val:
                return False

        return True

    def visit_Membership(self, expr, **_):
        return self.visit(expr.element) in set(self.visit(expr.set))

    def visit_RegexFilter(self, expr, **_):
        string = self.visit(expr.string)
        pattern = self.visit(expr.regex)

        return re.compile(pattern).match(str(string))

    def visit_StrictOrderedSet(self, expr, **_):
        iterator = iter(expr.children)
        min_ = self.visit(next(iterator))

        if min_ is None:
            return False

        for child in iterator:
            val = self.visit(child)

            if not min_ > val or val is None:
                return False

            min_ = val

        return True

    def visit_PartialOrderedSet(self, expr, **_):
        iterator = iter(expr.children)
        min_ = self.visit(next(iterator))

        if min_ is None:
            return False

        for child in iterator:
            val = self.visit(child)
            if min_ < val or val is None:
                return False

            min_ = val

        return True


engine.Engine.register_engine(ObjectMatcher, "filter")



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER rule-based query analyzer.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

from efilter import engine
from efilter import errors
from efilter import expression


class Hinter(engine.VisitorEngine):
    """Retargets the query to apply to a subexpression, to be used for hinting.

    This class follows the visitor pattern. See documentation on VisitorEngine.

    Discussion of mechanism and rationale:
    ======================================

    Bear with me on this one.

    As with any database-like system, certain EFILTER queries can be satisfied
    by following several equally valid, but differently expensive strategies.
    Because traditional notions of cardinality and available indexing do not
    apply to most systems that provide EFILTER with data (and, at any rate,
    such concerns are usually negligible in comparison with the actual
    collection of data), it becomes useful for the expert system supplying the
    data to be able to influence the strategy based on what it knows about the
    relative trade-offs.

    To give a concrete example, take the following query:
        VAD.flags contains {'execute', 'write'} and VAD.process.name == 'init'

    (In this example a VAD is a virtual address descriptor, which is a data
    structure used by operating systems to keep track of pageable memory.)

    A naive expert (uh...) system will generate all possible VADs and feed them
    to the EFILTER matcher engine. An optimized system - lets call it an expert
    expert system - can only collect VADs belonging to the 'init' process at
    fraction of the cost, and supply those to EFILTER for indexing and
    filtering. This is possible, because the expert system knows that processes
    are the first to be collected, and that each process holds a reference to
    related VADs, making it beneficial to skip processes that don't apply. What
    the expert system needs is a way to signal that such a strategy is
    available, and a way to recognize the processes which cannot be skipped.

    Enter the Hinter - taking the example above, the expert system can run:
        q = Query("VAD.flags contains {'execute', 'write'} "
                  " and VAD.process.name == 'init'")
        hint = q.run_engine('hinter', selector='VAD.process')

    This will cause the Hinter to generate a hint query equivalent to the
    expression "name == 'init'", which can be applied to a process object
    for prefiltering. Amazing.
    """

    def run(self, selector=None, *args, **kwargs):
        self.selector = tuple(selector.split(".")) if selector else ()

        # Make sure the query is reasonably shaped with respect to let-forms.
        self.query = self.query.run_engine("normalizer")

        expr = super(Hinter, self).run(*args, trace=(), **kwargs)
        return self.query.subquery(expr)

    def _next_crumb(self, trace):
        if self.selector[:len(trace)] == trace:
            next_crumb = self.selector[len(trace)]
            return next_crumb

        return None

    def visit_Let(self, expr, trace=None, **_):
        if self.selector == trace:
            # We're already in a subexpression we want to preserve.
            return expr

        self._last_let_form_reached_bottom = False

        if not isinstance(expr.lhs, expression.Binding):
            # Can't do anything - might not be correct - perhaps we should
            # blow up here? (Only Let-forms with a Binding on the LHS can be
            # hinted.)
            raise errors.EfilterError(
                query=self.query,
                message=("Hinter can only optimize let forms where lhs is a "
                         "binding (a variable). Got %r instead.")
                % (self.query.subquery(expr)),
                root=expr)

        crumb = expr.lhs.value
        next_crumb = self._next_crumb(trace)

        if next_crumb == crumb:
            # Descend.
            trace_ = trace + (next_crumb,)
            branch = self.visit(expr.rhs, trace=trace + (next_crumb,))

            if trace_ == self.selector:
                # Bottom.
                self._last_let_form_reached_bottom = True

            # This could have been set by the conditional above or by a nested
            # let-form. Either way, we keep the branch only if it matched the
            # entire selector, otherwise it's discarded.
            if self._last_let_form_reached_bottom:
                return branch

        return None  # Eliminate this branch.

    def _build_variadic(self, expr, **kwargs):
        children = []
        for child in expr.children:
            branch = self.visit(child, **kwargs)
            if branch:
                children.append(branch)

        return children

    def visit_VariadicExpression(self, expr, **kwargs):
        children = self._build_variadic(expr, **kwargs)
        if not children:
            return None

        if len(children) == 1:
            # We play it safe and keep the only child in case the return type
            # of this expression isn't just a boolean (math).
            # More specific visitors will eliminate the whole thing.
            return children[0]

        return type(expr)(*children)

    def visit_Relation(self, expr, **kwargs):
        children = self._build_variadic(expr, **kwargs)

        if len(children) == 1:
            return None  # It's pointless to execute this relation.

        return type(expr)(*children)

    def visit_ValueExpression(self, expr, **kwargs):
        child = self.visit(expr.value, **kwargs)
        if child is None:
            return None

        return type(expr)(child)

    def visit_ComponentLiteral(self, expr, **_):
        return expr

    def visit_BinaryExpression(self, expr, **kwargs):
        lhs = self.visit(expr.lhs, **kwargs)
        if not lhs:
            return None

        rhs = self.visit(expr.rhs, **kwargs)
        if not rhs:
            return None

        return type(expr)(lhs, rhs)

    def visit_Literal(self, expr, **_):
        return expr

    def visit_Binding(self, expr, trace=None, **_):
        if self.selector == trace:
            return expr

        return None


engine.Engine.register_engine(Hinter, "hinter")



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER abstract syntax.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

from efilter import protocol

from efilter.protocols import associative
from efilter.protocols import boolean
from efilter.protocols import eq
from efilter.protocols import iset
from efilter.protocols import ordered
from efilter.protocols import number


class Expression(object):
    """Base class of the query AST.

    Behavior of the query language is encoded in the various QueryVisitor
    subclasses. Expressions themselves only contain children and an arity hint.
    """

    __abstract = True
    children = ()
    arity = 0
    start = None
    end = None

    type_signature = (protocol.AnyType,)
    return_signature = protocol.AnyType

    def __hash__(self):
        return hash((type(self), self.children))

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.children == other.children

    def __ne__(self, other):
        return not self.__eq__(other)

    def __init__(self, *children, **kwargs):
        super(Expression, self).__init__()

        self.start = kwargs.pop("start", None)
        self.end = kwargs.pop("end", None)

        if kwargs:
            raise ValueError("Unexpected argument(s) %s" % kwargs.keys())

        if self.arity and len(children) != self.arity:
            raise ValueError("%d-ary expression %s passed %d children." % (
                self.arity, type(self).__name__, len(children)))

        self.children = children

    def __repr__(self):
        if len(self.children) == 1:
            return "%s(%r)" % (type(self).__name__, self.children[0])

        lines = []
        for child in self.children:
            if isinstance(child, Expression):
                clines = [" %s" % line for line in repr(child).split("\n")]
            else:
                clines = repr(child).split("\n")
            lines.extend(clines)

        return "%s(\n%s)" % (type(self).__name__, "\n".join(lines))


class ValueExpression(Expression):
    """Unary expression."""
    arity = 1

    @property
    def value(self):
        return self.children[0]


class BinaryExpression(Expression):
    arity = 2

    @property
    def lhs(self):
        return self.children[0]

    @property
    def rhs(self):
        return self.children[1]


class VariadicExpression(Expression):
    """Represents an expression with variable arity."""

    arity = None


### Value (unary) expressions ###

class Literal(ValueExpression):
    """Represents a literal, which is to say not-an-expression."""

    type_signature = None  # Depends on literal.


class Binding(ValueExpression):
    """Represents a member of the evaluated object - attributes of entity."""

    type_signature = (associative.IAssociative,)


class ComponentLiteral(ValueExpression):
    """Evaluates to True if the component exists."""


class Complement(ValueExpression):
    """Logical NOT."""

    type_signature = (boolean.IBoolean,)
    return_signature = boolean.IBoolean


### Binary expressions ###


class Let(BinaryExpression):
    """Let(BINDING, SUBQUERY) evaluates SUBQUERY with the result of BINDING.

    Example:
    # True if the parent of this process has a Timestamps component.
    Let("Process/parent", ComponentLiteral("Timestamps"))
    """

    type_signature = (associative.IAssociative, Expression)
    return_signature = None  # Depends on rhs.

    @property
    def context(self):
        return self.lhs

    @property
    def expression(self):
        return self.rhs


class LetAny(Let):
    """Like Let, but handles multiple BINDINGS using intersection semantics."""

    return_signature = boolean.IBoolean


class LetEach(Let):
    """Like Let, but handles multiple BINDINGS using union semantics."""

    return_signature = boolean.IBoolean


class Membership(BinaryExpression):
    """Membership of element in set."""
    type_signature = (eq.IEq, iset.ISet)
    return_signature = boolean.IBoolean

    @property
    def element(self):
        return self.lhs

    @property
    def set(self):
        return self.rhs


class RegexFilter(BinaryExpression):
    type_signature = (basestring, basestring)
    return_signature = boolean.IBoolean

    @property
    def string(self):
        return self.lhs

    @property
    def regex(self):
        return self.rhs


### Variadic Expressions ###

### Logical Variadic ###

class LogicalOperation(VariadicExpression):
    type_signature = boolean.IBoolean
    return_signature = boolean.IBoolean


class Union(LogicalOperation):
    """Logical OR (variadic)."""


class Intersection(LogicalOperation):
    """Logical AND (variadic)."""


### Variadic Relations ###

class Relation(VariadicExpression):
    return_signature = boolean.IBoolean


class OrderedSet(Relation):
    """Abstract class to represent strict and non-strict ordering."""

    type_signature = ordered.IOrdered


class StrictOrderedSet(OrderedSet):
    """Greater than relation."""

    type_signature = ordered.IOrdered


class PartialOrderedSet(OrderedSet):
    """Great-or-equal than relation."""

    type_signature = ordered.IOrdered


class ContainmentOrder(Relation):
    """Inclusion of set 1 by set 2 and so on."""

    type_signature = iset.ISet


class Equivalence(Relation):
    """Logical == (variadic)."""

    type_signature = eq.IEq


### Variadic Arithmetic ###

class NumericExpression(VariadicExpression):
    """Arithmetic expressions."""

    return_signature = number.INumber


class Sum(NumericExpression):
    """Arithmetic + (variadic)."""

    type_signature = number.INumber


class Difference(NumericExpression):
    """Arithmetic - (variadic)."""

    type_signature = number.INumber


class Product(NumericExpression):
    """Arithmetic * (variadic)."""

    type_signature = number.INumber


class Quotient(NumericExpression):
    """Arithmetic / (variadic)."""

    type_signature = number.INumber



# -*- coding: utf-8 -*-

# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""EFILTER abstract type system."""

from efilter import dispatch
from efilter import protocol

# Declarations:
# pylint: disable=unused-argument


@dispatch.polymorphic
def eq(x, y):
    raise NotImplementedError()


@dispatch.polymorphic
def ne(x, y):
    raise NotImplementedError()


class IEq(protocol.Protocol):
    _protocol_functions = (eq, ne)


# Default implementations:

IEq.implement(
    for_type=protocol.AnyType,
    implementations={
        eq: lambda x, y: x == y,
        ne: lambda x, y: x != y
    }
)
